__all__ = ()

from time import perf_counter

from scarletio import Task, copy_docs, from_json, repeat_timeout, sleep, to_json
from scarletio.web_common import ConnectionClosed, InvalidHandshake, URL, WebSocketProtocolError

from ..core import KOKORO

from .constants import (
    GATEWAY_ACTION_CONNECT, GATEWAY_ACTION_KEEP_GOING, GATEWAY_ACTION_RESUME, GATEWAY_CONNECT_TIMEOUT,
    GATEWAY_OPERATION_VOICE_CLIENT_CONNECT, GATEWAY_OPERATION_VOICE_CLIENT_DISCONNECT, GATEWAY_OPERATION_VOICE_FLAGS,
    GATEWAY_OPERATION_VOICE_HEARTBEAT, GATEWAY_OPERATION_VOICE_HEARTBEAT_ACKNOWLEDGE, GATEWAY_OPERATION_VOICE_HELLO,
    GATEWAY_OPERATION_VOICE_IDENTIFY, GATEWAY_OPERATION_VOICE_PLATFORM, GATEWAY_OPERATION_VOICE_READY,
    GATEWAY_OPERATION_VOICE_RESUME, GATEWAY_OPERATION_VOICE_RESUMED, GATEWAY_OPERATION_VOICE_SELECT_PROTOCOL,
    GATEWAY_OPERATION_VOICE_SESSION_DESCRIPTION, GATEWAY_OPERATION_VOICE_SPEAKING,
    GATEWAY_OPERATION_VOICE_USERS_CONNECT, GATEWAY_OPERATION_VOICE_VIDEO_SESSION_DESCRIPTION,
    GATEWAY_OPERATION_VOICE_VIDEO_SINK, LATENCY_DEFAULT, POLL_TIMEOUT
)
from .heartbeat import Kokoro
from .rate_limit import GatewayRateLimiter
from .voice_base import DiscordGatewayVoiceBase


async def _handle_operation_dummy(gateway, data):
    """
    Dummy operation handler.
    
    Parameters
    ----------
    gateway : ``DiscordGatewayVoice``
        The gateway receiving the operation.
    data : `None | dict<str, object>`
        The received data.
    
    Returns
    -------
    gateway_action : `int`
    """
    return GATEWAY_ACTION_KEEP_GOING


class DiscordGatewayVoice(DiscordGatewayVoiceBase):
    """
    The gateway used by ``VoiceClient``-s to communicate with Discord with secure web socket.
    
    Attributes
    ----------
    _operation_handlers : `dict<int, (instance, dict<str, object>) -> int>`
        Handler for each expected operation.
    
    _should_run : `bool`
        Whether the gateway should be running.
    
    kokoro : `None | Kokoro`
        The heart of the gateway, sends beat-data at set intervals. If does not receives answer in time, restarts
        the gateway.
    
    rate_limit_handler : ``GatewayRateLimiter``
        The rate limit handler of the gateway.
    
    sequence : `int`
        Last sequence number received.
    
    voice_client : ``VoiceClient``
        The owner voice client of the gateway.
    
    web_socket : `None`, ``WebSocketClient``
        The web socket of the gateway.
    """
    __slots__ = (
        '_operation_handlers', '_should_run', 'kokoro', 'rate_limit_handler', 'sequence', 'voice_client', 'web_socket'
    )
    
    def __new__(cls, voice_client):
        """
        Creates a voice gateway with it's default attributes.
        
        Parameters
        ----------
        voice_client : ``VoiceClient``
            The owner voice_client of the gateway.
        """
        operation_handlers = {
            # GATEWAY_OPERATION_VOICE_IDENTIFY: send only
            # GATEWAY_OPERATION_VOICE_SELECT_PROTOCOL: send only
            GATEWAY_OPERATION_VOICE_READY: cls._handle_operation_ready,
            # GATEWAY_OPERATION_VOICE_HEARTBEAT: send only
            GATEWAY_OPERATION_VOICE_SESSION_DESCRIPTION: cls._handle_operation_session_description,
            GATEWAY_OPERATION_VOICE_SPEAKING: cls._handle_operation_speaking,
            GATEWAY_OPERATION_VOICE_HEARTBEAT_ACKNOWLEDGE: cls._handle_operation_heartbeat_acknowledge,
            # GATEWAY_OPERATION_VOICE_RESUME: send only 
            GATEWAY_OPERATION_VOICE_HELLO: cls._handle_operation_hello,
            GATEWAY_OPERATION_VOICE_RESUMED: cls._handle_operation_resumed,
            GATEWAY_OPERATION_VOICE_USERS_CONNECT: cls._handle_operation_users_connect,
            GATEWAY_OPERATION_VOICE_CLIENT_CONNECT: cls._handle_operation_client_connect,
            GATEWAY_OPERATION_VOICE_CLIENT_DISCONNECT: cls._handle_operation_client_disconnect,
            GATEWAY_OPERATION_VOICE_VIDEO_SESSION_DESCRIPTION: _handle_operation_dummy,
            GATEWAY_OPERATION_VOICE_VIDEO_SINK: _handle_operation_dummy,
            GATEWAY_OPERATION_VOICE_FLAGS: _handle_operation_dummy,
            GATEWAY_OPERATION_VOICE_PLATFORM: _handle_operation_dummy,
        }
        
        self = object.__new__(cls)
        self._operation_handlers = operation_handlers
        self._should_run = False
        self.kokoro = None
        self.rate_limit_handler = GatewayRateLimiter()
        self.sequence = -1
        self.voice_client = voice_client
        self.web_socket = None
        return self
    
    
    def _create_kokoro(self):
        """
        Creates the gateway's ``.kokoro``.
        """
        kokoro = self.kokoro
        if kokoro is None:
            self.kokoro = Kokoro(self)
        else:
            kokoro.stop()
    
    
    @copy_docs(DiscordGatewayVoiceBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(' voice_client = ')
        repr_parts.append(repr(self.voice_client))

    
    @property
    @copy_docs(DiscordGatewayVoiceBase.latency)
    def latency(self):
        kokoro = self.kokoro
        if kokoro is None:
            latency = LATENCY_DEFAULT
        else:
            latency = kokoro.latency
        return latency
    
    
    @copy_docs(DiscordGatewayVoiceBase.send_as_json)
    async def send_as_json(self, data):
        web_socket = self.web_socket
        if web_socket is None:
            return
        
        if not (await self.rate_limit_handler):
            return
        
        try:
            await web_socket.send(to_json(data))
        except ConnectionClosed:
            pass
    
    
    @copy_docs(DiscordGatewayVoiceBase.beat)
    async def beat(self):
        data = {
            'op': GATEWAY_OPERATION_VOICE_HEARTBEAT,
            'd': {
                't': int(perf_counter() * 1000),
                'seq_ack': self.sequence,
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _voice_client_connect(self):
        """
        Sends a voice client connect packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.voice_client
        data = {
            'op': GATEWAY_OPERATION_VOICE_CLIENT_CONNECT,
            'd': {
                'audio_ssrc': voice_client._audio_source,
                'video_ssrc': voice_client._video_source,
                'rtx_ssrc': 0,
            }
        }
        
        await self.send_as_json(data)
    
    
    @copy_docs(DiscordGatewayVoiceBase.set_speaking)
    async def set_speaking(self, speaking):
        data = {
            'op': GATEWAY_OPERATION_VOICE_SPEAKING,
            'd': {
                'speaking': speaking,
                'delay': 0,
            },
        }
        
        await self.send_as_json(data)
    
    
    def _get_session_id(self):
        """
        Gets the session identifier of the connections.
        
        Returns
        -------
        session_id : `str`
        """
        voice_state = self.voice_client.voice_state
        if voice_state is None:
            session_id = ''
        else:
            session_id = voice_state.session_id
        
        return session_id
    
    
    async def _identify(self):
        """
        Sends an identify packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.voice_client
        
        data = {
            'op': GATEWAY_OPERATION_VOICE_IDENTIFY,
            'd': {
                'seq_ack': self.sequence,
                'server_id': str(voice_client.guild_id),
                'session_id': self._get_session_id(),
                'token': voice_client._token,
                'user_id': str(voice_client.client.id),
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _resume(self):
        """
        Sends a resume packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.voice_client
        
        data = {
            'op': GATEWAY_OPERATION_VOICE_RESUME,
            'd': {
                'seq_ack': self.sequence,
                'server_id': str(voice_client.guild_id),
                'session_id': self._get_session_id(),
                'token': voice_client._token,
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _select_protocol(self, ip, port):
        """
        Sends a select protocol packet to Discord.
        
        This method is a coroutine.
        
        Parameters
        ----------
        ip : `str`
            The received ip of the voice voice_client to use.
        port : `int`
            The received port of the voice voice_client to use.
        """
        data = {
            'op': GATEWAY_OPERATION_VOICE_SELECT_PROTOCOL,
            'd': {
                'protocol': 'udp',
                'data': {
                    'address': ip,
                    'port': port,
                    'mode': self.voice_client.get_encryption_mode(),
                },
            },
        }
        
        await self.send_as_json(data)
    
    
    def _cancel_self_and_get_web_socket(self):
        """
        Cancels the gateway except its web socket. Returns the web socket if still running instead.
        
        Returns
        -------
        web_socket : `None | WebSocketClient`
        """
        kokoro = self.kokoro
        if (kokoro is not None):
            kokoro.stop()
        
        web_socket = self.web_socket
        if web_socket is None:
            return None
        
        self.web_socket = None
        
        if web_socket.closed:
            return None
        
        return web_socket
    
    
    @copy_docs(DiscordGatewayVoiceBase.terminate)
    async def terminate(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            await web_socket.close(4000)
    
    
    @copy_docs(DiscordGatewayVoiceBase.close)
    async def close(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            await web_socket.close(1000)
    
    
    @copy_docs(DiscordGatewayVoiceBase.abort)
    def abort(self):
        web_socket = self._cancel_self_and_get_web_socket()
        if (web_socket is not None):
            web_socket.close_transport(True)
        
        self._should_run = False
    
    
    async def _connect(self, resume):
        """
        Connects the gateway to Discord. If the connecting was successful, will start it's `.kokoro` as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        resume : `bool`
            Whether the gateway should try to resume the existing connection.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionError
        OSError
        ValueError
        ConnectionClosed
        InvalidHandshake
        WebSocketProtocolError
        """
        self._create_kokoro()
        
        web_socket = self.web_socket
        if (web_socket is not None) and (not web_socket.closed):
            self.web_socket = None
            await web_socket.close(4000)
        
        gateway_url = URL(f'wss://{self.voice_client._endpoint}/?v=8', True)
        self.web_socket = await self.voice_client.client.http.connect_web_socket(gateway_url)
        self.kokoro.start()
        
        if not resume:
            await self._identify()
            return GATEWAY_ACTION_KEEP_GOING
        
        await self._resume()
        
        try:
            await self.web_socket.ensure_open()
        except ConnectionClosed:
            # web_socket got closed so let's just do a regular connect.
            self._clear_session()
            self.kokoro.stop()
            self.web_socket = None
            return GATEWAY_ACTION_CONNECT
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_users_connect(self, message):
        # Dont know what to do with the payload currently.
        # It looks like:
        # message = {'op': 11, 'd': {'user_ids': ['22221231231231231231']}}
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_client_connect(self, message):
        """
        Handles a voice client connect gateway operation.
        Note that not we are connecting, but someone else.
        
        This function is a coroutine.
        
        Parameters
        ----------
        data : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        user_id = int(data['user_id'])
        voice_client = self.voice_client
        
        try:
            audio_source = data['audio_ssrc']
        except KeyError:
            pass
        else:
            voice_client._update_audio_source(user_id, audio_source)
        
        try:
            video_source = data['video_ssrc']
        except KeyError:
            pass
        else:
            voice_client._update_video_source(user_id, video_source)
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_session_description(self, message):
        """
        Handles a voice session description operation.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        voice_client = self.voice_client
        voice_client.set_encryption_mode(data['mode'], bytes(data['secret_key']))
        
        await self.set_speaking(voice_client.speaking)
        
        voice_client.set_connected()
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_speaking(self, message):
        """
        Handles a speaking operation.
        Note that not we are speaking, but someone else.
        
        This function is a coroutine.
        
        Parameters
        ----------
        data : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        user_id = int(data['user_id'])
        audio_source = data['ssrc']
        self.voice_client._update_audio_source(user_id, audio_source)
        return GATEWAY_ACTION_KEEP_GOING

    
    async def _handle_operation_client_disconnect(self, message):
        """
        Handles a client disconnect operation.
        Note that not we are disconnecting, but someone else.
        
        This function is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        voice_client = self.voice_client
        
        user_id = int(data['user_id'])
        voice_client._remove_audio_source(user_id)
        voice_client._remove_video_source(user_id)
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_ready(self, message):
        """
        Handles a ready operation and selects protocol with `._select_protocol`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        data = message.get('d', None)
        if data is None:
            return GATEWAY_ACTION_KEEP_GOING
        
        kokoro = self.kokoro
        if (kokoro is None) or (kokoro.runner is None):
            return GATEWAY_ACTION_CONNECT
        
        # ignore setting heartbeat interval
        # kokoro.interval = data['heartbeat_interval'] / 1000.0
        
        voice_client = self.voice_client
        audio_source = data['ssrc']
        voice_client._audio_source = audio_source
        voice_client._endpoint_ip = data['ip']
        voice_client._endpoint_port = data['port']
        voice_client.prefer_encryption_mode_from_options(data['modes'])
        
        # data structure
        # byte  0 -  2 : `version` -> we send 1
        # byte  2 -  4 : `length` -> 70 (this excluded type and length)
        # byte  4 -  8 : `audio_source` value as unsigned integer
        # byte  8 - 72 : `ip` null terminated string, receive only but still have to send for some weird reason?
        # byte 72 - 74 : `port` as integer, receive only, but still have to send for some weird reason?
        packet = b''.join([b'\x00\x01\x00\x46', audio_source.to_bytes(4, 'big'), b'\x00' * 66])
        
        voice_client.send_packet(packet)
        
        protocol = voice_client._protocol
        # Make sure, that the voice voice client's reader do not wanna read our data away from us.
        protocol.cancel_current_reader()
        
        received = await protocol.read(74)
        
        ip_end = received.index(0, 8)
        if ip_end == -1:
            ip_end = 72
        elif ip_end > 72:
            ip_end = 72
        
        ip = received[8 : ip_end].decode('ascii')
        voice_client._ip = ip
        
        # port is last 2 bytes
        port = int.from_bytes(received[72 : 74], 'big')
        voice_client._port = port
        
        await self._select_protocol(ip, port)
        await self._voice_client_connect()
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_hello(self, message):
        """
        Handles a hello operation.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received data.
        
        Returns
        -------
        gateway_action : `int`
        """
        kokoro = self.kokoro
        if kokoro is None:
            kokoro = Kokoro(self)
            self.kokoro = kokoro
        
        # ignore setting heartbeat interval
        # kokoro.interval = data['heartbeat_interval'] / 1000.0
        
        # beat now calls `.start` since it has to check the current state
        await kokoro.beat_now()
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_heartbeat_acknowledge(self, message):
        """
        Handles a heart beat acknowledge operation.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        kokoro = self.kokoro
        if kokoro is None:
            kokoro = Kokoro(self)
            self.kokoro = kokoro
            # `.answered` is not calling `.start` 
            kokoro.start()
            
        kokoro.answered()
        
        return GATEWAY_ACTION_KEEP_GOING
    
    
    async def _handle_operation_resumed(self, message):
        """
        Handles a resumed operation. Called after the gateway was resumed successfully. We do nothing.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `dict<str, object>`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        return GATEWAY_ACTION_KEEP_GOING

    
    async def _handle_received_operation(self, message):
        """
        Processes the message sent by Discord.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `str`
            The received message.
        
        Returns
        -------
        gateway_action : `int`
        """
        message = from_json(message)
        
        sequence = message.get('seq', None)
        if (sequence is not None):
            self.sequence = sequence
        
        operation = message['op']
        
        try:
            operation_handler = self._operation_handlers[operation]
        except KeyError:
            client = self.voice_client.client
            Task(
                KOKORO,
                client.events.error(
                    client,
                    f'{type(self).__name__}._handle_received_operation',
                    f'Unknown operation: {operation!r}\nMessage: {message!r}'
                ),
            )
            return GATEWAY_ACTION_KEEP_GOING
        
        return (await operation_handler(self, message))
    
    
    async def _poll_and_handle_received_operation(self):
        """
        Waits to receive a message from Discord, then calls ``._handle_received_operation``.
        
        This method is a coroutine.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionClosed
            If the web_socket connection closed.
        """
        web_socket = self.web_socket
        if web_socket is None:
            return GATEWAY_ACTION_CONNECT
        
        message = await web_socket.receive()
        return await self._handle_received_operation(message)
    
    
    async def _keep_polling_and_handling(self):
        """
        Keeps polling and handling till decided otherwise.
        
        This function is a coroutine.
        
        Returns
        -------
        gateway_action : `int`
        
        Raises
        ------
        ConnectionClosed
            If the web_socket connection closed.
        """
        try:
            with repeat_timeout(POLL_TIMEOUT) as loop:
                for _ in loop:
                    action = await self._poll_and_handle_received_operation()
                    if action != GATEWAY_ACTION_KEEP_GOING:
                        return action
        
        except TimeoutError:
            # timeout, no internet probably.
            pass
        
        return GATEWAY_ACTION_CONNECT
    
    
    @copy_docs(DiscordGatewayVoiceBase.run)
    async def run(self, waiter = None):
        self._should_run = True
        voice_client = self.voice_client
        action = GATEWAY_ACTION_CONNECT
        
        try:
            while True:
                voice_client.clear_connected()
                if (not self._should_run) or (not voice_client.running):
                    return False
                
                try:
                    task = Task(KOKORO, self._connect(action == GATEWAY_ACTION_RESUME))
                    task.apply_timeout(GATEWAY_CONNECT_TIMEOUT)
                    action = await task
                    
                    if action != GATEWAY_ACTION_KEEP_GOING:
                        continue
                    
                    if (waiter is not None):
                        waiter.set_result_if_pending(True)
                        waiter = None
                    
                    action = await self._keep_polling_and_handling()
                
                except GeneratorExit:
                    self.abort()
                    return False
                
                except TimeoutError:          
                    pass
                
                except (ConnectionError, InvalidHandshake, OSError, ValueError, WebSocketProtocolError) as exception:
                    from scarletio import write_exception_async
                    await write_exception_async(exception)
                    
                    # We are not connected anymore, clear connection waiter.
                    voice_client.clear_connected()
                    await sleep(1.0, KOKORO)
                
                else:
                    continue
        
        except:
            raise
        
        finally:
            if (waiter is not None):
                waiter.set_result_if_pending(False)
                waiter = None
            
            # we are not running anymore.
            self._should_run = False
        
        return False
    
    
    def _clear_session(self):
        """
        Clears current session data, disabling the option of resuming the connection.
        """
        self.sequence = -1



__all__ = ()

from itertools import chain

from ...bases.entity.slotted_meta import _get_direct_parent, _inherit_hash, _merge_type_slots

from ..core import DEFAULT_EVENT_HANDLER

from .helpers import _merge_to_type
from .event import Event
from .event_deprecation_descriptor import EventDeprecationDescriptor


def _get_reserved_attributes_from_parent(meta_type, direct_parent, type_parents):
    """
    Gets the reserved attributes from the parent type.
    
    Parameters
    ----------
    meta_type : `type`
        The meta-type.
    
    direct_parent : `type`
        The directly inherited type.
    
    type_parents : `tuple<type>`
        Cumulative parent types.
    
    Returns
    -------
    parent_event_names : `None | frozenset<str>`
    parent_default_handlers : `None | tuple<(str, async-callable, bool)>`
    parent_parameter_counts : `None | dict<str, int>`
    parent_event_deprecations : `None | frozenset<(str, EventDeprecation)>
    
    Raises
    ------
    RuntimeError
        - If a non `meta_type` parent type implements any reserved attribute.
    """
    if isinstance(direct_parent, meta_type):
        parent_event_names = direct_parent._plugin_event_names
        parent_default_handlers = direct_parent._plugin_default_handlers
        parent_parameter_counts = direct_parent._plugin_parameter_counts
        parent_event_deprecations = direct_parent._plugin_event_deprecations
    
    else:
        for type_parent in type_parents:
            for attribute_name in (
                '_plugin_event_names',
                '_plugin_default_handlers',
                '_plugin_parameter_counts',
                '_plugin_event_deprecations',
            ):
                if (getattr(type_parent, attribute_name, None) is not None):
                    raise RuntimeError(
                        f'Non {meta_type.__name__} parent type shall not implement `{attribute_name!r}`, '
                        f'but {type_parent!r} does.'
                    )
        
        parent_event_names = None
        parent_default_handlers = None
        parent_parameter_counts = None
        parent_event_deprecations = None
    
    return parent_event_names, parent_default_handlers, parent_parameter_counts, parent_event_deprecations


def _separate_events(type_attributes):
    """
    Separates out the events from the type's attributes.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        Type attributes.
    
    Returns
    -------
    events : `list<(str, Event)>`
    events_deprecated : `list<(str, Event)>`
    """
    # Collect
    events = []
    events_deprecated = []
    
    for type_attribute_name, type_attribute_value in type_attributes.items():
        if not isinstance(type_attribute_value, Event):
            continue
        
        if type_attribute_value.deprecation is None:
            container = events
        else:
            container = events_deprecated
        
        container.append((type_attribute_name, type_attribute_value))
    
    # Separate
    for event_name, event in chain(events, events_deprecated):
        del type_attributes[event_name]
    
    return events, events_deprecated


def _finalize_slots(events, slots):
    """
    Updates the slots with.
    
    Parameters
    ----------
    events : `list<(str, Event)>`
        Events to collect extra slots from.
    
    slots : `set<str>`
        Already collected slots.
    
    Returns
    -------
    finalized_sots : `tuple<str>`
    """
    slots.update(event_name for event_name, event in events)
    return tuple(sorted(slots))


def _collect_event_names(events, parent_event_names):
    """
    Collects the event names.
    
    Parameters
    ----------
    events : `list<(str, Event)>`
        Events to collect from.
    
    parent_event_names : `None | frozenset<str>`
        Event names of the parent type.
    
    Returns
    -------
    event_names : `None | frozenset<str>`
    """
    event_names = None
    for event_name, event in events:
        if event_names is None:
            event_names = []
        
        event_names.append(event_name)
    
    return _merge_to_type(parent_event_names, event_names, frozenset)


def _collect_default_handlers(events, parent_default_handlers):
    """
    Collects the default handler for each event.
    
    Parameters
    ----------
    events : `list<(str, Event)>`
        Events to collect from.
    
    parent_default_handlers : `None | tuple<(str, async-callable, bool)>`
        Default handlers of the parent type.
    
    Returns
    -------
    default_handlers : `None | tuple<(str, async-callable, bool)>`
    """
    default_handlers = None
    
    for event_name, event in events:
        default_handler = event.default_handler
        if (default_handler is None):
            default_handler = DEFAULT_EVENT_HANDLER
            instance_default_handler = False
        
        else:
            instance_default_handler = event.instance_default_handler
        
        if default_handlers is None:
            default_handlers = []
        
        default_handlers.append((event_name, default_handler, instance_default_handler))
    
    return _merge_to_type(parent_default_handlers, default_handlers, tuple)


def _collect_parameter_counts(events, parent_parameter_counts):
    """
    Collects the parameter counts for each event.
    
    Parameters
    ----------
    events : `list<(str, Event)>`
        Events to collect from.
    
    parent_parameter_counts : `None | dict<str, int>`
        Parameter counts of the parent type.
    
    Returns
    -------
    parameter_counts : `None | dict<str, int>`
    """
    parameter_counts = None
    
    for event_name, event in events:
        if (parameter_counts is None):
            parameter_counts = {}
        
        parameter_counts[event_name] = event.parameter_count
    
    if (parent_parameter_counts is not None):
        if (parameter_counts is None):
            parameter_counts = {}
        
        parameter_counts.update(parent_parameter_counts)
    
    return parameter_counts


def _collect_event_deprecations(events_deprecated, parent_event_deprecations):
    """
    Collects the deprecations from each event.
    
    Parameters
    ----------
    events_deprecated : `list<(str, Event)>`
        Deprecated events.
    
    parent_event_deprecations : `frozenset<(str, EventDeprecation)>`
        Deprecated events of the parent type.
    
    Returns
    -------
    event_deprecations : `frozenset<(str, EventDeprecation)>`
    """
    return _merge_to_type(
        (((event_name, event.deprecation) for event_name, event in events_deprecated) if events_deprecated else None),
        parent_event_deprecations,
        frozenset,
    )


class EventHandlerPluginMeta(type):
    """
    Meta type type of ``EventHandlerPlugin`` sub-types.
    
    Contains the critical logic to collect and after lookup the details about the added events.
    """
    def __new__(cls, type_name, type_parents, type_attributes):
        """
        Creates a new ``Slotted``.
        
        Parameters
        ----------
        type_name : `str`
            The created type's name.
        
        type_parents : `tuple<type>`
            The parent types of the creates type.
        
        type_attributes : `dict<str, object>`
            The type attributes of the created type.
        
        Raises
        ------
        RuntimeError
            - If there is a duplicate event entry.
        """
        direct_parent = _get_direct_parent(cls, type_name, type_parents)
        final_slots = _merge_type_slots(direct_parent, type_parents, type_attributes)
        _inherit_hash(direct_parent, type_attributes)
        
        (
            parent_event_names,
            parent_default_handlers,
            parent_parameter_counts,
            parent_event_deprecations,
        ) = _get_reserved_attributes_from_parent(cls, direct_parent, type_parents)
        events, events_deprecated = _separate_events(type_attributes)
        type_attributes['_plugin_event_names'] = _collect_event_names(events, parent_event_names)
        type_attributes['_plugin_default_handlers'] = _collect_default_handlers(events, parent_default_handlers)
        type_attributes['_plugin_parameter_counts'] = _collect_parameter_counts(events, parent_parameter_counts)
        type_attributes['_plugin_event_deprecations'] = _collect_event_deprecations(
            events_deprecated, parent_event_deprecations
        )
        type_attributes['__slots__'] = _finalize_slots(events, final_slots)
        type_attributes.update(
            (event_name, EventDeprecationDescriptor(event_name, event.deprecation))
            for event_name, event in events_deprecated
        )
        
        return type.__new__(cls, type_name, type_parents, type_attributes)
    
    
    def __call__(cls, *positional_parameters, **keyword_parameters):
        """
        Instances the type.
        
        Auto-sets event handlers.
        
        Parameters
        ----------
        *positional_parameters : Positional parameters
            Additional positional parameters.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Returns
        -------
        instance : `instance<cls>`
        """
        instance = cls.__new__(cls, *positional_parameters, **keyword_parameters)
        if type(instance) is cls:
            default_handlers = cls._plugin_default_handlers
            if (default_handlers is not None):
                for event_name, default_handler, instance_default_handler in default_handlers:
                    if instance_default_handler:
                        default_handler = default_handler()
                    
                    object.__setattr__(instance, event_name, default_handler)
            
                
            cls.__init__(instance, *positional_parameters, **keyword_parameters)
        
        return instance



__all__ = ('DiscordException',)

import re

from scarletio.web_common.headers import DATE, RETRY_AFTER

from ...env import RICH_DISCORD_EXCEPTION

from ..utils import DATETIME_FORMAT_CODE, parse_date_header_to_datetime


if RICH_DISCORD_EXCEPTION:
    from .payload_renderer import reconstruct_payload
else:
    reconstruct_payload = None


EXCEPTION_RESPONSE_RP = re.compile('\\d+\\: (.*)')


class DiscordException(Exception):
    r"""
    Represents an exception raised by Discord, when it response with a not expected response code.
    
    Depending on Discord's response code, the http client's behaviours differently.
    
    +-------+-----------------------+---------------+
    | Code  | Meaning               | Behaviour     |
    +=======+=======================+===============+
    | 200   | OK                    | return        |
    +-------+-----------------------+---------------+
    | 201   | CREATED               | return        |
    +-------+-----------------------+---------------+
    | 204   | NO CONTENT            | return        |
    +-------+-----------------------+---------------+
    | 304   | NOT MODIFIED          | return        |
    +-------+-----------------------+---------------+
    | 400   | BAD REQUEST           | raise         |
    +-------+-----------------------+---------------+
    | 401   | UNAUTHORIZED          | raise         |
    +-------+-----------------------+---------------+
    | 403   | FORBIDDEN             | raise         |
    +-------+-----------------------+---------------+
    | 404   | NOT FOUND             | raise         |
    +-------+-----------------------+---------------+
    | 405   | METHOD NOT ALLOWED    | raise         |
    +-------+-----------------------+---------------+
    | 429   | TOO MANY REQUESTS     | rate limited  |
    +-------+-----------------------+---------------+
    | 500   | SERVER ERROR          | \*retry       |
    +-------+-----------------------+---------------+
    | 502   | GATEWAY UNAVAILABLE   | \*retry       |
    +-------+-----------------------+---------------+
    | 5XX   | SERVER ERROR          | raise         |
    +-------+-----------------------+---------------+
    
    \* For five times a request can fail with `OsError` or return `501` / `502` response code. If the request fails
    with these cases for the fifth try and the last one resulted `501` / `502` response code, then
    ``DiscordException`` will be raised.
    
    Attributes
    ----------
    _code : `None | int`
        Cache of the `.code` property.
        
        If the response data does not contains `code`, then this attribute is set to `0`.
    
    _debug_info : `None | str`
        cache of the `.debug_info` property.
    
    _errors : `None | list<str>`
        Cache of the `errors` property.
    
    _request_info : `None | str`
        Cache of the `.request_info` property.
    
    _retry_after : `None | float`
        Cache pf the `.retry_after` property.
    
    _message : `None | str`
        Cache of the `.message` property.
    
    _status : `None | int`
        Cache of the `.status` property.
    
    debug_options : `None | set<str>`
        Debug options of the http client.
    
    received_data : `object`
        Deserialized `json` response data if applicable.
    
    response : ``ClientResponse``
        The http client response, what caused the error.
    
    sent_data : `object`
        Sent data.
    """
    __slots__ = (
        '_code', '_debug_info', '_errors', '_request_info', '_retry_after', '_message', '_status', 'debug_options',
        'received_data', 'response', 'sent_data',
    )
    
    def __new__(cls, response, received_data, sent_data, debug_options):
        """
        Creates a new discord exception.
        
        Parameters
        ----------
        response : `None | ClientResponse`
            The http client response, what caused the error.
        received_data : `object`
            Deserialized `json` response data if applicable.
        sent_data : `object`
            Sent data.
        debug_options : `None | set<str>`
            Debug options of the http client.
        """
        self = Exception.__new__(cls, response, received_data, sent_data, debug_options)
        self._code = None
        self._debug_info = None
        self._errors = None
        self._request_info = None
        self._retry_after = None
        self._message = None
        self._status = None
        self.debug_options = debug_options
        self.received_data = received_data
        self.response = response
        self.sent_data = sent_data
        return self
    
    
    __init__ = object.__init__
    
    
    def _get_message(self):
        """
        Creates the error message of the discord exception.
        
        Returns
        -------
        message : `str`
        """
        message_parts = []
        
        message_parts.append(type(self).__name__)
        response = self.response
        if (response is not None):
            message_parts.append(' ')
            message_parts.append(response.reason)
            message_parts.append(' (')
            message_parts.append(repr(response.status))
            message_parts.append(')')
        
        code = self.code
        if code:
            message_parts.append(', code = ')
            message_parts.append(repr(code))
        
        received_data = self.received_data
        if isinstance(received_data, dict):
            message_base = received_data.get('message', '')
        else:
            message_base = ''
        
        if message_base:
            # At some cases we might get message like `400: text` At this case want to cut the status part.
            matched = EXCEPTION_RESPONSE_RP.fullmatch(message_base)
            if (matched is not None):
                message_base = matched.group(1)
            
            message_parts.append(': ')
            message_parts.append(message_base)
        
        
        if self.status == 429:
            if not message_base.endswith(('.', ',')):
                message_parts.append(';')
            
            message_parts.append(' retry after: ')
            message_parts.append(format(self.retry_after, '.02f'))
        
        return ''.join(message_parts)
    
    
    @property
    def message(self):
        """
        Returns the error message of the discord exception.
        
        Returns
        -------
        message : `str`
        """
        message = self._message
        if message is None:
            message = self._get_message()
            self._message = message
        
        return message
    
    
    @message.setter
    def message(self, value):
        self._message = value
    
    
    @message.deleter
    def message(self):
        self._message = None
    
    
    def _get_request_info(self):
        """
        Creates additional request information about the exception.
        
        Returns
        -------
        request_info : `str`
        """
        response = self.response
        if response is None:
            return ''
        
        message_parts = []
        
        message_parts.append(format(parse_date_header_to_datetime(response.headers[DATE]), DATETIME_FORMAT_CODE))
        message_parts.append('; ')
        message_parts.append(response.method)
        message_parts.append(' ')
        message_parts.append(str(response.url))
        
        return ''.join(message_parts)
    
    
    @property
    def request_info(self):
        """
        Returns additional request information about the exception.
        
        Returns
        -------
        request_info : `str`
        """
        request_info = self._request_info
        if request_info is None:
            request_info = self._get_request_info()
            self._request_info = request_info
        
        return request_info
    
    
    @request_info.setter
    def request_info(self, value):
        self._request_info = value
    
    
    @request_info.deleter
    def request_info(self):
        self._request_info = None
    
    
    def _get_errors(self):
        """
        Creates the errors shipped with the exception.
        
        Might return an empty list.
        
        Returns
        -------
        errors : `list` of `str`
        """
        messages = []
        
        message_parts = []
        received_data = self.received_data
        if isinstance(received_data, dict):
            error_datas = received_data.get('errors', None)
            if (error_datas is not None) and error_datas:
                stack = [[(None, error_datas)]]
                while True:
                    line = stack[-1]
                    if not line:
                        del stack[-1]
                        if not stack:
                            break
                            
                        del stack[-1][-1]
                        if not message_parts:
                            continue
                            
                        del message_parts[-1]
                        if not message_parts:
                            continue
                        
                        if message_parts[-1] != '.':
                            continue
                        
                        del message_parts[-1]
                        continue
                    
                    key, value = line[-1]
                    
                    if isinstance(value, dict):
                        if (key is not None):
                            if key.isdigit():
                                # this should not be first ever
                                message_parts.append(f'[{key}]')
                            else:
                                if message_parts:
                                    message_parts.append('.')
                                message_parts.append(key)
                        try:
                            error_datas = value['_errors']
                        except KeyError:
                            stack.append([*value.items()])
                            continue
                        
                        for error_data in error_datas:
                            error_data_length = len(error_data)
                            try:
                                error_code = error_data['code']
                            except KeyError:
                                error_code = 'ERROR'
                            else:
                                error_data_length -= 1
                            
                            try:
                                error_message = error_data['message']
                            except KeyError:
                                error_message = None
                            else:
                                error_data_length -= 1
                            
                            if error_data_length:
                                error_extra = ' '.join(
                                    f'{key} = {value!r}' for key, value in error_data.items()
                                    if key not in ('code', 'message')
                                )
                                
                                if (error_message is None):
                                    error_message = error_extra
                                else:
                                    error_message = f'{error_message!r} {error_extra}'
                            
                            elif (error_message is None):
                                error_message = ''
                            else:
                                error_message = repr(error_message)
                            
                            if message_parts:
                                message_parts.append('.')
                            
                            message_parts.append(f'{error_code}({error_message})')
                            messages.append(''.join(message_parts))
                            del message_parts[-1]
                            if not message_parts:
                                continue
                            
                            if message_parts[-1] != '.':
                                continue
                            
                            del message_parts[-1]
                            continue
                        
                        del line[-1]
                        if not message_parts:
                            continue
                            
                        del message_parts[-1]
                        
                        if not message_parts:
                            continue
                            
                        if message_parts[-1] != '.':
                            continue
                        
                        del message_parts[-1]
                        continue
                    
                    if key.isdigit():
                        message_parts.append(f'[{key}]')
                    else:
                        message_parts.append('.')
                        message_parts.append(key)
                    message_parts.append('.')
                    message_parts.append(value)
                    messages.append(''.join(message_parts))
                    del line[-1]
                    del message_parts[-3:]
                    if not message_parts:
                        continue
                    
                    if message_parts[-1] != '.':
                        continue
                    
                    del message_parts[-1]
        
        messages.reverse()
        return messages
    
    
    @property
    def errors(self):
        """
        Returns the errors shipped with the exception.
        
        Might return an empty list.
        
        Returns
        -------
        errors : `list` of `str`
        """
        errors = self._errors
        if errors is None:
            errors = self._get_errors()
            self._errors = errors
        
        return errors
    
    
    @errors.setter
    def errors(self, value):
        self._errors = value
    
    
    @errors.deleter
    def errors(self):
        self._errors = None
    
    
    def _create_debug_info(self):
        """
        Creates debug info.
        
        Returns
        -------
        debug_info : `str`
        """
        debug_options = self.debug_options
        if (debug_options is None):
            return ''
        
        debug_info_parts = ['debug options: ']
        
        debug_options = sorted(debug_options)
        
        index = 0
        limit = len(debug_options)
        while True:
            debug_option = debug_options[index]
            index += 1
            
            debug_info_parts.append(debug_option)
            
            if index == limit:
                break
            
            debug_info_parts.append(', ')
            continue
        
        return ''.join(debug_info_parts)
    
    
    @property
    def debug_info(self):
        """
        Returns debug info.
        
        Only displayed within `.messages` when rich discord exceptions are enabled.
        
        Returns
        -------
        debug_info : `str`
        """
        
        debug_info = self._debug_info
        if debug_info is None:
            debug_info = self._create_debug_info()
            self._debug_info = debug_info
        
        return debug_info
    
    
    @debug_info.setter
    def debug_info(self, value):
        self._debug_info = value
    
    
    @debug_info.deleter
    def debug_info(self):
        self._debug_info = None
    
    
    @property
    def messages(self):
        """
        Returns a list of the errors. The 0th element of the list is always a header line, what contains the
        exception's name, the response's reason and it's status. If set, then also the Discord's internal error code
        and it's message as well.
        
        Every other element at the list is optional. Those are extra errors included in the response's data.
        
        Returns
        -------
        messages : `list` of `str`
        """
        messages = []
        
        messages.append(self.message)
        messages.append(self.request_info)
        
        messages.extend(self.errors)
        
        if RICH_DISCORD_EXCEPTION:
            debug_info = self.debug_info
            if debug_info:
                messages.append(debug_info)
            
            reconstructed_payload = reconstruct_payload(self.sent_data)
            if (reconstructed_payload is not None):
                messages.append(reconstructed_payload)
        
        return messages
    
    
    @messages.setter
    def messages(self, value):
        self._messages = value
    
    
    @messages.deleter
    def messages(self):
        self._messages = None
    
    
    def _get_code(self):
        """
        Parses out the Discord's inner exception code from the response's data. Sets it to `._code` and returns it as
        well.
        
        Returns
        -------
        error_code : `int`
        """
        received_data = self.received_data
        if isinstance(received_data, dict):
            code = received_data.get('code', 0)
        else:
            code = 0
        
        return code
    
    
    @property
    def code(self):
        """
        Returns the Discord's internal exception code, if it is included in the response's data. If not, then returns
        `0`.
        
        Returns
        -------
        error_code : `int`
        """
        code = self._code
        if code is None:
            code = self._get_code()
            self._code = code
        
        return code
    
    
    @code.setter
    def code(self, value):
        self._code = value
    
    
    @code.deleter
    def code(self):
        self._code = None
    
    
    def _get_status(self):
        """
        Returns the response's status. Defaults to `0`.
        
        Returns
        -------
        status : `int`
        """
        response = self.response
        if (response is None):
            return 0
        
        return response.status
        
    
    @property
    def status(self):
        """
        The exception's response's status.
        
        Returns
        -------
        status : `int`
        """
        status = self._status
        if status is None:
            status = self._get_status()
            self._status = status
        
        return status
    
    
    @status.setter
    def status(self, value):
        self._status = value
    
    
    @status.deleter
    def status(self):
        self._status = None
    
    
    def _get_retry_after(self):
        """
        Returns after how much seconds the request should be retried. Defaults to `0.0`.
        
        Returns
        -------
        retry_after : `float`
        """
        response = self.response
        if response is None:
            return 0.0
        
        try:
            retry_after = response.headers[RETRY_AFTER]
        except KeyError:
            return 0.0
        
        try:
            retry_after = float(retry_after)
        except ValueError:
            return 0.0
        
        return retry_after
    
    
    @property
    def retry_after(self):
        """
        After how much seconds the request should be retried.
        Applicable for rate limit errors, so the ones with status `429`.
        
        Returns
        -------
        retry_after : `float`
        """
        retry_after = self._retry_after
        if retry_after is None:
            retry_after = self._get_retry_after()
            self._retry_after = retry_after
        
        return retry_after
    
    
    @retry_after.setter
    def retry_after(self, value):
        self._retry_after = value
    
    
    @retry_after.deleter
    def retry_after(self):
        self._retry_after = None
    
    
    def __repr__(self):
        """Returns the representation of the object."""
        return '\n'.join(self.messages)
    
    
    __str__ = __repr__
    
    
    def __format__(self, code):
        """
        Formats the discord exception in a format string.
        
        The only provided option is no code, which returns a short-ish representation.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        exception : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        """
        if not code:
            return f'<{self.message}>'
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}.'
        )



__all__ = ('MessageInteraction',)

from scarletio import export, include

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ZEROUSER, create_partial_user_from_id

from .fields import (
    parse_authorizer_user_ids, parse_id, parse_interacted_message_id, parse_name_and_sub_command_name_stack,
    parse_response_message_id, parse_target_message_id, parse_target_user, parse_triggering_interaction, parse_type,
    parse_user, put_authorizer_user_ids, put_id, put_interacted_message_id,
    put_name_and_sub_command_name_stack, put_response_message_id, put_target_message_id,
    put_target_user, put_triggering_interaction, put_type, put_user, validate_authorizer_user_ids,
    validate_id, validate_interacted_message_id, validate_name, validate_response_message_id,
    validate_sub_command_name_stack, validate_target_message_id, validate_target_user, validate_triggering_interaction,
    validate_type, validate_user
)


InteractionType = include('InteractionType')


PRECREATE_FIELDS = {
    'authorizer_users': ('authorizer_user_ids', validate_authorizer_user_ids),
    'authorizer_user_ids': ('authorizer_user_ids', validate_authorizer_user_ids),
    'interacted_message': ('interacted_message_id', validate_interacted_message_id),
    'interacted_message_id': ('interacted_message_id', validate_interacted_message_id),
    'interaction_type': ('type', validate_type),
    'name': ('name', validate_name),
    'response_message': ('response_message_id', validate_response_message_id),
    'response_message_id': ('response_message_id', validate_response_message_id),
    'sub_command_name_stack': ('sub_command_name_stack', validate_sub_command_name_stack),
    'target_message': ('target_message_id', validate_target_message_id),
    'target_message_id': ('target_message_id', validate_target_message_id),
    'target_user': ('target_user', validate_target_user),
    'triggering_interaction': ('triggering_interaction', validate_triggering_interaction),
    'user': ('user', validate_user),
}


@export   
class MessageInteraction(DiscordEntity):
    """
    Sent with a ``Message`` when the it is a response to an ``InteractionEvent``.
    
    Attributes
    ----------
    authorizer_user_ids : `None | dict<ApplicationIntegrationType, int>`
        The users' identifier who authorized the integration.
    
    id : `int`
        The interaction's identifier.
    
    interacted_message_id : `int`
        The interacted message's identifier. Present if the message is created from a component interaction.
    
    name : `str`
        The invoked interaction's name.
    
    response_message_id : `int`
        The response message's identifier. Present if the message is a followup one.
    
    sub_command_name_stack : `None | tuple<str>`
        The sub-command-group and sub-command names.
    
    target_message_id : `int`
        The targeted message's identifier in case of a message context interaction.
    
    target_user : `None | ClientUserBase`
        The targeted message's identifier in case of a user context interaction.
    
    triggering_interaction : `None | MessageInteraction`
        Represents the source form interaction. Present if the message is created from a form interaction.
    
    type : ``InteractionType``
        The interaction's type.
    
    user : ``ClientUserBase``
        The user who invoked the interaction.
    """
    __slots__ = (
        'authorizer_user_ids', 'interacted_message_id', 'name', 'response_message_id', 'sub_command_name_stack',
        'target_message_id', 'target_user', 'triggering_interaction', 'type', 'user'
    )
    
    def __new__(
        cls,
        *,
        authorizer_user_ids = ...,
        interacted_message_id = ...,
        interaction_type = ...,
        name = ...,
        response_message_id = ...,
        sub_command_name_stack = ...,
        target_message_id = ...,
        target_user = ...,
        triggering_interaction = ...,
        user = ...,
    ):
        """
        Creates a new message interaction with the given fields.
        
        Parameters
        ----------
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        interacted_message_id : ``None | int | Message``, Optional (Keyword only)
            The interacted message's identifier. Present if the message is created from a component interaction.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        response_message_id : ``None | int | Message``, Optional (Keyword only)
            The response message's identifier. Present if the message is a followup one.
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        target_message_id : `int | Message`, Optional (Keyword only)
            The targeted message's identifier in case of a message context interaction.
        
        target_user : `None | ClientUserBase`, Optional (Keyword only)
            The targeted message's identifier in case of a user context interaction.
        
        triggering_interaction : `None | MessageInteraction`, Optional (Keyword only)
            Represents the source form interaction. Present if the message is created from a form interaction.
        
        user : ``None | ClientUserBase``, Optional (Keyword only)
            The user who invoked the interaction.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # authorizer_user_ids
        if authorizer_user_ids is ...:
            authorizer_user_ids = None
        else:
            authorizer_user_ids = validate_authorizer_user_ids(authorizer_user_ids)
        
        # interacted_message_id
        if interacted_message_id is ...:
            interacted_message_id = 0
        else:
            interacted_message_id = validate_interacted_message_id(interacted_message_id)
        
        # interaction_type
        if interaction_type is ...:
            interaction_type = InteractionType.none
        else:
            interaction_type = validate_type(interaction_type)
        
        # response_message_id
        if response_message_id is ...:
            response_message_id = 0
        else:
            response_message_id = validate_response_message_id(response_message_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # sub_command_name_stack
        if sub_command_name_stack is ...:
            sub_command_name_stack = None
        else:
            sub_command_name_stack = validate_sub_command_name_stack(sub_command_name_stack)
        
        # target_message_id
        if target_message_id is ...:
            target_message_id = 0
        else:
            target_message_id = validate_target_message_id(target_message_id)
        
        # target_user
        if target_user is ...:
            target_user = None
        else:
            target_user = validate_target_user(target_user)
        
        # triggering_interaction
        if triggering_interaction is ...:
            triggering_interaction = None
        else:
            triggering_interaction = validate_triggering_interaction(triggering_interaction)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.authorizer_user_ids = authorizer_user_ids
        self.id = 0
        self.interacted_message_id = interacted_message_id
        self.name = name
        self.response_message_id = response_message_id
        self.sub_command_name_stack = sub_command_name_stack
        self.target_message_id = target_message_id
        self.target_user = target_user
        self.triggering_interaction = triggering_interaction
        self.type = interaction_type
        self.user = user
        return self
    
    
    @classmethod
    def precreate(cls, message_interaction_id, **keyword_parameters):
        """
        Creates a new message interaction with the given predefined fields.
        
        > Since message interactions are not globally cached, this method is only used for testing.
        
        Parameters
        ----------
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        authorizer_users : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            Alternative for `authorizer_user_ids`.
        
        message_interaction_id : `int`
            The message interaction's id.
        
        **keyword_parameters : Keyword parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        interacted_message : ``None | int | Message``, Optional (Keyword only)
            Alternative for `interacted_message_id`.
        
        interacted_message_id : ``None | int | Message``, Optional (Keyword only)
            The interacted message's identifier. Present if the message is created from a component interaction.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        response_message : ``None | int | Message``, Optional (Keyword only)
            Alternative for `response_message_id`.
        
        response_message_id : ``None | int | Message``, Optional (Keyword only)
            The response message's identifier. Present if the message is a followup one.
        
        target_message : `int | Message`, Optional (Keyword only)
            Alternative for `target_message_id`.
        
        target_message_id : `int | Message`, Optional (Keyword only)
            The targeted message's identifier in case of a message context interaction.
        
        target_user : `None | ClientUserBase`, Optional (Keyword only)
            The targeted message's identifier in case of a user context interaction.
        
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        triggering_interaction : `None | MessageInteraction`, Optional (Keyword only)
            Represents the source form interaction. Present if the message is created from a form interaction.
        
        user : ``None | ClientUserBase``, Optional (Keyword only)
            The user who invoked the interaction.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        message_interaction_id = validate_id(message_interaction_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(message_interaction_id)
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, message_interaction_id):
        """
        Creates a new message interaction with it's defaults attributes set.
        
        Parameters
        ----------
        message_interaction_id : `int`
            The message interaction's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.authorizer_user_ids = None
        self.id = message_interaction_id
        self.interacted_message_id = 0
        self.name = ''
        self.response_message_id = 0
        self.sub_command_name_stack = None
        self.target_message_id = 0
        self.target_user = None
        self.triggering_interaction = None
        self.type = InteractionType.none
        self.user = ZEROUSER
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new message interaction from the received data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Message interaction data.
        
        Returns
        -------
        data : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.authorizer_user_ids = parse_authorizer_user_ids(data)
        self.id = parse_id(data)
        self.interacted_message_id = parse_interacted_message_id(data)
        self.name, self.sub_command_name_stack = parse_name_and_sub_command_name_stack(data)
        self.response_message_id = parse_response_message_id(data)
        self.target_message_id = parse_target_message_id(data)
        self.target_user = parse_target_user(data)
        self.triggering_interaction = parse_triggering_interaction(data)
        self.type = parse_type(data)
        self.user = parse_user(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Tries to convert the message interaction back to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_authorizer_user_ids(self.authorizer_user_ids, data, defaults)
        put_interacted_message_id(self.interacted_message_id, data, defaults)
        put_name_and_sub_command_name_stack((self.name, self.sub_command_name_stack), data, defaults)
        put_response_message_id(self.response_message_id, data, defaults)
        put_target_message_id(self.target_message_id, data, defaults)
        put_target_user(self.target_user, data, defaults)
        put_triggering_interaction(
            self.triggering_interaction, data, defaults, include_internals = include_internals
        )
        put_type(self.type, data, defaults)
        
        if include_internals:
            put_user(self.user, data, defaults)
            put_id(self.id, data, defaults)
        
        return data
    
    
    @property
    def partial(self):
        """
        Returns whether the message interaction is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    
    def __repr__(self):
        """Returns the message interaction's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # id
        interaction_id = self.id
        if interaction_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(interaction_id))
            repr_parts.append(',')
        else:
            repr_parts.append(' (partial)')
            
        # type
        interaction_type = self.type
        repr_parts.append(' type = ')
        repr_parts.append(interaction_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(interaction_type.value))
        
        # name
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        # sub_command_name_stack
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            repr_parts.append(', sub_command_name_stack = ')
            repr_parts.append(repr(sub_command_name_stack))
        
        # user
        repr_parts.append(', user = ')
        repr_parts.append(repr(self.user))
        
        # authorizer_user_ids
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            repr_parts.append(', authorizer_user_ids = ')
            repr_parts.append(repr(authorizer_user_ids))
        
        # interacted_message_id
        interacted_message_id = self.interacted_message_id
        if (interacted_message_id is not None):
            repr_parts.append(', interacted_message_id = ')
            repr_parts.append(repr(interacted_message_id))
        
        # response_message_id
        response_message_id = self.response_message_id
        if response_message_id:
            repr_parts.append(', response_message_id = ')
            repr_parts.append(repr(response_message_id))
        
        # target_message_id
        target_message_id = self.target_message_id
        if target_message_id:
            repr_parts.append(', target_message_id = ')
            repr_parts.append(repr(target_message_id))
        
        # target_user
        target_user = self.target_user
        if target_user:
            repr_parts.append(', target_user = ')
            repr_parts.append(repr(target_user))
        
        # triggering_interaction
        triggering_interaction = self.triggering_interaction
        if (triggering_interaction is not None):
            repr_parts.append(', triggering_interaction = ')
            repr_parts.append(repr(triggering_interaction))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two message interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two message interactions not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id and self_id != other_id:
            return False
        
        # authorizer_user_ids
        if self.authorizer_user_ids != other.authorizer_user_ids:
            return False
        
        # interacted_message_id
        if self.interacted_message_id != other.interacted_message_id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # response_message_id
        if self.response_message_id != other.response_message_id:
            return False
        
        # sub_command_name_stack
        if self.sub_command_name_stack != other.sub_command_name_stack:
            return False
        
        # target_message_id
        if self.target_message_id != other.target_message_id:
            return False
        
        # target_user
        if self.target_user is not other.target_user:
            return False
        
        # triggering_interaction
        if self.triggering_interaction != other.triggering_interaction:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # user
        if self.user is not other.user:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the message integration's hash value."""
        hash_value = 0
        
        # authorizer_user_ids
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            hash_value ^= len(authorizer_user_ids) << 8
            for integration_type, user_id in authorizer_user_ids.items():
                hash_value ^= (integration_type.value << 22) & user_id
        
        # id
        hash_value ^= self.id
        
        # interacted_message_id
        hash_value ^= self.interacted_message_id
        
        # name
        hash_value ^= hash(self.name)
        
        # response_message_id
        hash_value ^= self.response_message_id
        
        # sub_command_name_stack
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            hash_value ^= len(sub_command_name_stack)
            for sub_command_name in sub_command_name_stack:
                hash_value ^= hash(sub_command_name)
        
        # target_message_id
        hash_value ^= self.target_message_id
        
        # target_user
        target_user = self.target_user
        if (target_user is not None):
            hash_value ^= target_user.id
        
        # triggering_interaction
        triggering_interaction = self.triggering_interaction
        if (triggering_interaction is not None):
            hash_value ^= hash(triggering_interaction)
        
        # type
        hash_value ^= self.type.value << 4
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    
    def copy(self):
        """
        Copies the message application returning a new partial one.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            authorizer_user_ids = authorizer_user_ids.copy()
        new.authorizer_user_ids = authorizer_user_ids
        
        new.id = 0
        new.interacted_message_id = self.interacted_message_id
        new.name = self.name
        new.response_message_id = self.response_message_id
        
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            sub_command_name_stack = (*sub_command_name_stack,)
        new.sub_command_name_stack = sub_command_name_stack
        
        new.target_message_id = self.target_message_id
        new.target_user = self.target_user
        
        triggering_interaction = self.triggering_interaction
        if (triggering_interaction is not None):
            triggering_interaction = triggering_interaction.copy()
        new.triggering_interaction = triggering_interaction
        
        new.type = self.type
        new.user = self.user
        
        return new
    
    
    def copy_with(
        self,
        *,
        authorizer_user_ids = ...,
        interacted_message_id = ...,
        interaction_type = ...,
        name = ...,
        response_message_id = ...,
        sub_command_name_stack = ...,
        target_message_id = ...,
        target_user = ...,
        triggering_interaction = ...,
        user = ...,
    ):
        """
        Copies the message interaction with the given fields returning a new partial one.
        
        Parameters
        ----------
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        interacted_message_id : ``None | int | Message``, Optional (Keyword only)
            The interacted message's identifier. Present if the message is created from a component interaction.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        name : `str`, Optional (Keyword only)
            The invoked interaction's name.
        
        response_message_id : ``None | int | Message``, Optional (Keyword only)
            The response message's identifier. Present if the message is a followup one.
        
        sub_command_name_stack : `None`, `iterable` of `str`, Optional (Keyword only)
            The sub-command-group and sub-command names.
        
        target_message_id : `int | Message`, Optional (Keyword only)
            The targeted message's identifier in case of a message context interaction.
        
        target_user : `None | ClientUserBase`, Optional (Keyword only)
            The targeted message's identifier in case of a user context interaction.
        
        triggering_interaction : `None | MessageInteraction`, Optional (Keyword only)
            Represents the source form interaction. Present if the message is created from a form interaction.
        
        user : ``None | ClientUserBase``, Optional (Keyword only)
            The user who invoked the interaction.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # authorizer_user_ids
        if authorizer_user_ids is ...:
            authorizer_user_ids = self.authorizer_user_ids
            if (authorizer_user_ids is not None):
                authorizer_user_ids = authorizer_user_ids.copy()
        else:
            authorizer_user_ids = validate_authorizer_user_ids(authorizer_user_ids)
        
        # interacted_message_id
        if interacted_message_id is ...:
            interacted_message_id = self.interacted_message_id
        else:
            interacted_message_id = validate_interacted_message_id(interacted_message_id)
        
        # interaction_type
        if interaction_type is ...:
            interaction_type = self.type
        else:
            interaction_type = validate_type(interaction_type)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # response_message_id
        if response_message_id is ...:
            response_message_id = self.response_message_id
        else:
            response_message_id = validate_response_message_id(response_message_id)
        
        # sub_command_name_stack
        if sub_command_name_stack is ...:
            sub_command_name_stack = self.sub_command_name_stack
            if (sub_command_name_stack is not None):
                sub_command_name_stack = (*sub_command_name_stack,)
        else:
            sub_command_name_stack = validate_sub_command_name_stack(sub_command_name_stack)
        
        # target_message_id
        if target_message_id is ...:
            target_message_id = self.target_message_id
        else:
            target_message_id = validate_target_message_id(target_message_id)
        
        # target_user
        if target_user is ...:
            target_user = self.target_user
        else:
            target_user = validate_target_user(target_user)
        
        # triggering_interaction
        if triggering_interaction is ...:
            triggering_interaction = self.triggering_interaction
            if (triggering_interaction is not None):
                triggering_interaction = triggering_interaction.copy()
        else:
            triggering_interaction = validate_triggering_interaction(triggering_interaction)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
    
        # Construct
        new = object.__new__(type(self))
        new.authorizer_user_ids = authorizer_user_ids
        new.id = 0
        new.interacted_message_id = interacted_message_id
        new.name = name
        new.response_message_id = response_message_id
        new.sub_command_name_stack = sub_command_name_stack
        new.target_message_id = target_message_id
        new.target_user = target_user
        new.triggering_interaction = triggering_interaction
        new.type = interaction_type
        new.user = user
        return new
    
    
    @property
    def joined_name(self):
        """
        Returns the joined name of the message interaction.
        
        Returns
        -------
        joined_name : `str`
        """
        name = self.name
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is None):
            return name
        
        return ' '.join([name, *sub_command_name_stack])
    
    
    @property
    def user_id(self):
        """
        Returns the user's identifier who invoked the interaction.
        
        Returns
        -------
        user_id : `int`
        """
        return self.user.id
    
    
    def get_authorizer_user_id(self, integration_type):
        """
        Gets the authorizer user's identifier for the given integration type.
        
        Parameters
        ----------
        integration_type : `ApplicationIntegrationType | int`
            Integration type to query for.
        
        Returns
        -------
        user_id : `int`
        """
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            try:
                return authorizer_user_ids[integration_type]
            except KeyError:
                pass
        
        return 0
    
    
    def get_authorizer_user(self, integration_type):
        """
        Gets the authorizer user for the given integration type.
        
        Parameters
        ----------
        integration_type : `ApplicationIntegrationType | int`
            Integration type to query for.
        
        Returns
        -------
        user : `None | ClientUserBase`
        """
        user_id = self.get_authorizer_user_id(integration_type)
        if user_id:
            return create_partial_user_from_id(user_id)

    
    @property
    def target_user_id(self):
        """
        Returns the targeted user's identifier.
        
        Returns
        -------
        target_user_id : `int`
        """
        target_user = self.target_user
        if target_user is None:
            return 0
        
        return target_user.id



__all__ = ('create_auto_custom_id', 'create_identifier_custom_id_from_name')

from base64 import b85encode as to_base85
from os import urandom as random_bytes


def create_auto_custom_id():
    """
    Creates a random custom identifier for components.
    
    Returns
    -------
    custom_id : `str`
        The created custom id.
    """
    return to_base85(random_bytes(64)).decode()


def create_identifier_custom_id_from_name(name):
    """
    Tries to turn the given name into an identifier. The identifier is always lower case. Spaces and minus signs are
    turned to underscore.
    
    Parameters
    ----------
    name : `str`
        The name of a custom_id.
    
    Returns
    -------
    custom_id : `str`
        The created custom id.
    """
    return name.casefold().replace(' ', '_').replace('-', '_')



__all__ = ()

from ...field_parsers import flag_parser_factory, nullable_flag_parser_factory
from ...field_putters import flag_optional_putter_factory, nullable_flag_optional_putter_factory
from ...field_validators import flag_validator_factory, nullable_flag_validator_factory
from ...color import Color

# primary

parse_color_primary = flag_parser_factory('primary_color', Color)
put_color_primary = flag_optional_putter_factory('primary_color', Color())
validate_color_primary = flag_validator_factory('color_primary', Color)


# secondary

parse_color_secondary = nullable_flag_parser_factory('secondary_color', Color)
put_color_secondary = nullable_flag_optional_putter_factory('secondary_color')
validate_color_secondary = nullable_flag_validator_factory('color_secondary', Color)


# tertiary

parse_color_tertiary = nullable_flag_parser_factory('tertiary_color', Color)
put_color_tertiary = nullable_flag_optional_putter_factory('tertiary_color')
validate_color_tertiary = nullable_flag_validator_factory('color_tertiary', Color)



__all__ = ()

from ...field_parsers import int_parser_factory, nullable_entity_parser_factory
from ...field_putters import int_putter_factory, nullable_entity_optional_putter_factory
from ...field_validators import int_conditional_validator_factory, nullable_entity_validator_factory

from ..sku_enhancement_guild import SKUEnhancementGuild


# boost_cost

parse_boost_cost = int_parser_factory('boost_price', 0)
put_boost_cost = int_putter_factory('boost_price')
validate_boost_cost = int_conditional_validator_factory(
    'boost_cost',
    0,
    (lambda boost_cost : boost_cost >= 0),
    '>= 0',
)


# guild

parse_guild = nullable_entity_parser_factory('guild_features', SKUEnhancementGuild)
put_guild = nullable_entity_optional_putter_factory('guild_features', SKUEnhancementGuild)
validate_guild = nullable_entity_validator_factory('guild', SKUEnhancementGuild)


# purchase_limit

parse_purchase_limit = int_parser_factory('purchase_limit', 0)
put_purchase_limit = int_putter_factory('purchase_limit')
validate_purchase_limit = int_conditional_validator_factory(
    'purchase_limit',
    0,
    (lambda purchase_limit : purchase_limit >= 0),
    '>= 0',
)



__all__ = ()

from ...field_parsers import entity_id_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import entity_id_putter_factory, force_string_putter_factory, nullable_string_putter_factory
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_string_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

from .constants import DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN

# bot

def parse_bot(data):
    """
    Parses the integration application's bot from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Integration application data.
    
    Returns
    -------
    bot : ``ClientUserBase``
    """
    user_data = data.get('bot', None)
    if user_data is None:
        bot = ZEROUSER
    else:
        bot = User.from_data(user_data)
    
    return bot


def put_bot(user, data, defaults, *, include_internals = False):
    """
    Puts the `bot`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The integration application's bot.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    include_internals : `bool` = `False`, Optional (Keyword only)
        Whether internal fields should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (user is not ZEROUSER):
        if user is ZEROUSER:
            user_data = None
        
        else:
            # User not implements `defaults` and `include_internals` for now, so we wont forward them.
            user_data = user.to_data(defaults = defaults, include_internals = include_internals)
        
        data['bot'] = user_data
    
    return data        


def validate_bot(value):
    """
    Validates whether the given value is an application integration bot.
    
    Parameters
    ----------
    value : ``None | ClientUserBase``
        The integration application bot.
    
    Returns
    -------
    bot : ``ClientUserBase``
    
    Raises
    ------
    TypeError
        - If `value` is not ``ClientUserBase``.
    """
    if value is None:
        bot = ZEROUSER
    
    elif isinstance(value, ClientUserBase):
        bot = value
    
    else:
        raise TypeError(
            f'`bot` can be `None`, `{ClientUserBase.__name__}`, got {value.__class__.__name__}; {value!r}.'
        )
    
    return bot

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('integration_application_id')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ('ComponentMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder

from .constants import SEPARATOR_SPACING_SIZE_DEFAULT
from .preinstanced import ButtonStyle, TextInputStyle


class ComponentMetadataBase(RichAttributeErrorBaseType):
    """
    Base type for component metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new component metadata from the given fields.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new component metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to build the metadata from.
        
        Returns
        -------
        self : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a keyword parameter's type is incorrect.
        ValueError
            - If a keyword parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the component metadata's representation."""
        return f'<{type(self).__name__}>'
    
    
    def __hash__(self):
        """Returns the component metadata's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two component metadatas are equal."""
        if type(self) is not type(other):
            return False
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. Other must be the same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new component metadata from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Component data.
        
        Returns
        -------
        self : `instance<cls>`
            Returns the created component.
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Returns the component metadata's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        return {}
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the component metadata by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : ``None | Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy(self):
        """
        Copies the component metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the component metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy()
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the component metadata with the given keyword parameters.
        
        The used up fields are popped from the keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters defining which fields should be changed.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy_with()
    
    
    def iter_contents(self):
        """
        Iterates over the contents of the component metadata.
        
        This method is an iterable generator.
        
        Yields
        ------
        content : `str`
        """
        return
        yield
    
    
    # Field placeholders
    
    button_style = PlaceHolder(
        ButtonStyle.none,
        """
        The component's style. Applicable for button components.
        
        Returns
        -------
        button_style : ``ButtonStyle``
        """
    )
    
    
    channel_types = PlaceHolder(
        None,
        """
        The allowed channel types by the select.
        
        Returns
        -------
        channel_types : ``None | tuple<ChannelType>``
        """
    )
    
    
    color = PlaceHolder(
        None,
        """
        The color of the component. At the case of container components it means the strip on the left.
        Passing `0` means black.
        
        color : ``None | Color``
        """
    )
    
    
    component = PlaceHolder(
        None,
        """
        The sub-component nested inside (the label component).
        
        Returns
        -------
        component : ``None | Component``
        """
    )
    
    
    components = PlaceHolder(
        None,
        """
        Sub-components.
        
        Returns
        -------
        components : ``None | tuple<Component>``
        """
    )
    
    
    content = PlaceHolder(
        None,
        """
        The content shown on the component.
        
        Returns
        -------
        content : `None | str`
        """
    )
    
    
    custom_id = PlaceHolder(
        None,
        """
        Custom identifier to detect which component was clicked (or used) by the user.
        
        > Mutually exclusive with the `sku_id` and `url` fields if the component is a buttons.
        
        Returns
        -------
        custom_id : `None | str`
        """
    )
    
    
    default_values = PlaceHolder(
        None,
        """
        Entities suggested by default in an entity select.
        
        Returns
        -------
        default_values : ``None | tuple<EntitySelectDefaultValue>``
        """
    )
    
    
    description = PlaceHolder(
        None,
        """
        Description of the component's media.
        
        Returns
        -------
        description : `None | str`
        """
    )
    
    
    divider = PlaceHolder(
        True,
        """
        Whether the separator should contain a divider.
        
        Returns
        -------
        divider : `bool`
        """
    )
    
    
    emoji = PlaceHolder(
        None,
        """
        Emoji of the component if applicable.
        
        Returns
        -------
        emoji : ``None | Emoji``
        """
    )
    
    
    enabled = PlaceHolder(
        True,
        """
        Whether the component is enabled.
        
        Returns
        -------
        enabled : `bool`
        """
    )
    
    
    items = PlaceHolder(
        None,
        """
        The media items shown on the component.
        
        Returns
        -------
        items : ``None | tuple<MediaItem>``
        """
    )
    
    
    label = PlaceHolder(
        None,
        """
        Label of the component.
        
        Returns
        -------
        label : `None | str`
        """
    )
    
    
    max_length = PlaceHolder(
        0,
        """
        The maximal length of the inputted text.
        
        Returns
        -------
        max_length : `int`
        """
    )
    
    
    max_values = PlaceHolder(
        1,
        """
        The maximal amount of options to select.
        
        Returns
        -------
        max_values : `int`
        """
    )
    
    
    media = PlaceHolder(
        None,
        """
        The media of the component.
        
        Returns
        -------
        media : ``None | MediaInfo``
        """
    )
    
    
    min_length = PlaceHolder(
        0,
        """
        The minimal length of the inputted text.
        
        Returns
        -------
        min_length : `int`
        """
    )
    
    
    min_values = PlaceHolder(
        1,
        """
        The minimal amount of options to select.
        
        Returns
        -------
        min_values : `int`
        """
    )
    
    name = PlaceHolder(
        '',
        """
        The name of the component's attachment media.
        
        Returns
        -------
        name : `str`
        """
    )
    
    options = PlaceHolder(
        None,
        """
        Options of the component.
        
        Returns
        -------
        options : ``None | tuple<StringSelectOption>``
        """
    )
    
    
    placeholder = PlaceHolder(
        None,
        """
        Place holder text of the select.
        
        Returns
        -------
        placeholder : `None | str`
        """
    )
    
    
    required = PlaceHolder(
        False,
        """
        Whether the field is required to be fulfilled.
        
        Returns
        -------
        required : `bool`
        """
    )
    
    size = PlaceHolder(
        0,
        """
        The size of the component's attachment media.
        
        Returns
        -------
        size : `int`
        """
    )
    
    
    sku_id = PlaceHolder(
        0,
        """
        Purchasable stock keeping unit identifier.
        
        > Mutually exclusive with the `custom_id` and `url` fields if the component is a buttons.
        
        Returns
        -------
        sku_id : `int`
        """
    )
    
    
    spacing_size = PlaceHolder(
        SEPARATOR_SPACING_SIZE_DEFAULT,
        """
        The separator's spacing's size.
        
        Returns
        -------
        spacing_size : ``SeparatorSpacingSize``
        """
    )
    
    
    spoiler = PlaceHolder(
        False,
        """
        Whether the media or the content of the component is spoilered.
        
        Returns
        -------
        spoiler : `bool`
        """
    )
    
    
    text_input_style = PlaceHolder(
        TextInputStyle.none,
        """
        The text input's style.
        
        Returns
        -------
        text_input_style : ``TextInputStyle``
        """
    )
    
    
    thumbnail = PlaceHolder(
        None,
        """
        The thumbnail or other accessory (button) of a section component.
        
        Returns
        -------
        thumbnail : ``None | Component``
        """
    )
    
    
    url = PlaceHolder(
        None,
        """
        Url to redirect to.
         
        > Mutually exclusive with the `custom_id` and `sku_id` fields if the component is a buttons.
        
        Returns
        -------
        url : `None | str`
        """
    )
    
    
    value = PlaceHolder(
        None,
        """
        The input component's default value.
        
        Returns
        -------
        value : `None | str`
        """
    )



from .attachment_input import *
from .attachment_media import *
from .base import *
from .button import *
from .channel_select import *
from .constants import *
from .container import *
from .fields import *
from .label import *
from .media_gallery import *
from .mentionable_select import *
from .preinstanced import *
from .role_select import *
from .row import *
from .section import *
from .select_base import *
from .separator import *
from .string_select import *
from .text_display import *
from .text_input import *
from .thumbnail_media import *
from .user_select import *


__all__ = (
    *attachment_input.__all__,
    *attachment_media.__all__,
    *base.__all__,
    *button.__all__,
    *channel_select.__all__,
    *constants.__all__,
    *container.__all__,
    *fields.__all__,
    *label.__all__,
    *media_gallery.__all__,
    *mentionable_select.__all__,
    *preinstanced.__all__,
    *role_select.__all__,
    *row.__all__,
    *section.__all__,
    *select_base.__all__,
    *separator.__all__,
    *string_select.__all__,
    *text_display.__all__,
    *text_input.__all__,
    *thumbnail_media.__all__,
    *user_select.__all__,
)



__all__ = ()

from ..resolved import Resolved

from .resolver import Resolver


def _string_resolve_single(resolved, value):
    """
    String resolver for single value.
    
    Parameters
    ----------
    resolved : ``None | Resolved``
        Resolved to use.
    
    value : `None | str`
        Value to resolve.
    
    Returns
    -------
    value : `None | str`
    """
    return value


def _string_iter_resolve_single(resolved, value):
    """
    Iterative string resolver for single value.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    resolved : ``None | Resolved``
        Resolved to use.
    
    value : `None | str`
        Value to resolve.
    
    Yields
    -------
    value : `str`
    """
    if value is None:
        return
    
    yield value


def _string_resolve_multiple(resolved, values):
    """
    String resolver for multiple values.
    
    Parameters
    ----------
    resolved : ``None | Resolved``
        Resolved to use.
    
    values : `None | tuple<str>`
        Values to resolve.
    
    Returns
    -------
    values : `None | tuple<str>`
    """
    return values


def _iter_resolve_string_multiple(resolved, values):
    """
    Iterative string resolver for multiple values.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    resolved : ``None | Resolved``
        Resolved to use.
    
    values : `None | tuple<str>`
        Values to resolve.
    
    Yields
    -------
    value : `str`
    """
    if values is None:
        return
    
    yield from values


def _entity_resolver_factory(resolver_function):
    """
    Creates resolver functions with the given resolver function. 
    
    Parameters
    ----------
    resolver_function : ``FunctionType``
        Resolver function to use.
    
    Returns
    -------
    resolver_functions : `(FunctionType, FunctionType, FunctionType, FunctionType)`
    """
    def _entity_resolve_single(resolved, value):
        """
        Entity resolver for single value.
        
        Parameters
        ----------
        resolved : ``None | Resolved``
            Resolved to use.
        
        value : `None | str`
            Value to resolve.
        
        Returns
        -------
        value : `None | resolver_function.return`
        """
        nonlocal resolver_function
        
        if (resolved is None):
            return
        
        if (value is None):
            return
        
        try:
            entity_id = int(value)
        except ValueError:
            return
        
        return resolver_function(resolved, entity_id)
    
    
    def _entity_iter_resolve_single(resolved, value):
        """
        Iterative entity resolver for single value.
        
        This function is an iterable generator.
        
        Parameters
        ----------
        resolved : ``None | Resolved``
            Resolved to use.
        
        value : `None | str`
            Value to resolve.
        
        Yields
        -------
        value : `resolver_function.return`
        """
        nonlocal resolver_function
        
        if (resolved is None):
            return
        
        if (value is None):
            return
        
        try:
            entity_id = int(value)
        except ValueError:
            return
        
        entity = resolver_function(resolved, entity_id)
        if entity is None:
            return
        
        yield entity
    
    
    def _entity_resolve_multiple(resolved, values):
        """
        Entity resolver for multiple values.
        
        Parameters
        ----------
        resolved : ``None | Resolved``
            Resolved to use.
        
        values : `None | tuple<str>`
            Values to resolve.
        
        Returns
        -------
        values : `None | tuple<resolver_function.return>`
        """
        nonlocal resolver_function
        
        if (resolved is None):
            return
        
        if (values is None):
            return
        
        entities = None
        
        for value in values:
            try:
                entity_id = int(value)
            except ValueError:
                continue
            
            entity = resolver_function(resolved, entity_id)
            if entity is None:
                continue
            
            if entities is None:
                entities = []
            
            entities.append(entity)
            continue
        
        if (entities is not None):
            entities = tuple(entities)
        
        return entities
    
    
    def _entity_iter_resolve_multiple(resolved, values):
        """
        Iterative entity resolver for multiple values.
        
        This function is an iterable generator.
        
        Parameters
        ----------
        resolved : ``None | Resolved``
            Resolved to use.
        
        values : `None | tuple<str>`
            Values to resolve.
        
        Yields
        -------
        value : `resolver_function.return`
        """
        nonlocal resolver_function
        
        if (resolved is None):
            return
        
        if (values is None):
            return
        
        for value in values:
            try:
                entity_id = int(value)
            except ValueError:
                continue
            
            entity = resolver_function(resolved, entity_id)
            if entity is None:
                continue
            
            yield entity
            continue
    
    return _entity_resolve_single, _entity_iter_resolve_single, _entity_resolve_multiple, _entity_iter_resolve_multiple


RESOLVER_STRING = Resolver(
    'string',
    _string_resolve_single,
    _string_iter_resolve_single,
    _string_resolve_multiple,
    _iter_resolve_string_multiple,
)


RESOLVER_ATTACHMENT = Resolver(
    'attachment',
    *_entity_resolver_factory(Resolved.resolve_attachment),
)

RESOLVER_CHANNEL = Resolver(
    'channel',
    *_entity_resolver_factory(Resolved.resolve_channel),
)

RESOLVER_ROLE = Resolver(
    'role',
    *_entity_resolver_factory(Resolved.resolve_role),
)

RESOLVER_MENTIONABLE = Resolver(
    'mentionable',
    *_entity_resolver_factory(Resolved.resolve_mentionable),
)

RESOLVER_MESSAGE = Resolver(
    'message',
    *_entity_resolver_factory(Resolved.resolve_message),
)

RESOLVER_USER = Resolver(
    'user',
    *_entity_resolver_factory(Resolved.resolve_user),
)



__all__ = ('PluginTree',)

from itertools import chain

from scarletio import RichAttributeErrorBaseType

from ..plugin import Plugin

from .plugin_helpers import _sort_plugins, _unwrap_plugin


class PluginTree(RichAttributeErrorBaseType):
    """
    Represents a standalone plugin tree.
    
    Attributes
    ----------
    _plugins : `frozenset` of ``Plugin``
        Unordered plugins of the plugin tree.
    _plugins_sorted : `None`, `list` of ``Plugin``
        Cache fields for ``.get_plugins_sorted``.
    """
    __slots__ = ('_plugins', '_plugins_sorted')
    
    def __new__(cls, plugin, deep):
        """
        Creates a new plugin tree instance.
        
        Parameters
        ----------
        plugin : ``Plugin``
            Plugin to create tree from.
        deep : `bool`
            Whether the plugin with all of it's parent and with their child should be returned.
        """
        plugins = frozenset(_unwrap_plugin(plugin, deep))
        
        self = object.__new__(cls)
        self._plugins = plugins
        self._plugins_sorted = None
        return self
    
    
    def __repr__(self):
        """Returns the plugin tree's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        try:
            entry_plugin = self.get_entry_plugin()
        except RuntimeError:
            pass
        else:
            repr_parts.append(' entry_plugin = ')
            repr_parts.append(repr(entry_plugin.name))
            repr_parts.append(',')
        
        repr_parts.append(' plugin_count = ')
        repr_parts.append(repr(self.get_plugin_count()))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the hash value of the plugin tree."""
        return hash(self._plugins)
    
    
    def __eq__(self, other):
        """Returns whether the two plugin trees are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._plugins == other._plugins
    
    
    def __gt__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        return self.get_entry_plugin() > other.get_entry_plugin()
    
    
    
    def get_plugins_sorted(self):
        """
        Returns the sorted plugins.
        
        Returns
        -------
        plugins : `list` of ``Plugin``
        """
        plugins_sorted = self._plugins_sorted
        if plugins_sorted is None:
            plugins_sorted = _sort_plugins(self._plugins)
            self._plugins_sorted = plugins_sorted
        
        return plugins_sorted
    
    
    def get_entry_plugin(self):
        """
        Gets the entry plugin.
        
        Returns
        -------
        plugin : ``Plugin``
        """
        return self.get_plugins_sorted()[-1]
    
    
    def get_plugin_count(self):
        """
        Returns how much plugins are in the plugin tree.
        
        Returns
        -------
        plugin_count : `int`
        """
        return len(self._plugins)
    
    
    def has_plugin(self, plugin):
        """
        Returns whether the plugin tree has the given plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to check.
        
        Returns
        -------
        has_plugin : `bool`
        """
        return plugin in self._plugins
    
    
    def has_plugin_tree(self, plugin_tree):
        """
        Returns whether the plugin tree is a superset of the given plugin tree.
        
        Parameters
        ----------
        plugin_tree : `instance<type<self>>`
            The plugin tree to check.
        
        Returns
        -------
        is_superset : `bool`
        """
        return plugin_tree._plugins.issuperset(self._plugins)
    
    
    def is_equal_to_plugin(self, plugin):
        """
        Returns whether the plugin tree equals to the given plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to check.
        
        Returns
        -------
        is_equal : `bool`
        """
        plugins = self._plugins
        return len(plugins) == 1 and next(iter(plugins)) is plugin
    
    
    def is_equal_to_plugin_tree(self, plugin_tree):
        """
        Returns whether the plugin tree equals to the given plugin tree.
        
        Parameters
        ----------
        plugin_tree : `instance<type<self>>`
            The plugin tree to check.
        
        Returns
        -------
        is_equal : `bool`
        """
        return (self._plugins == plugin_tree._plugins)
    
    
    def is_subset_of_plugin(self, plugin):
        """
        Returns whether the plugin tree is a subset of the given plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to check.
        
        Returns
        -------
        is_subset : `bool`
        """
        return (not self._plugins)
    
    
    def is_subset_of_plugin_tree(self, plugin_tree):
        """
        Returns whether the plugin tree is a sub set the given plugin tree.
        
        Parameters
        ----------
        plugin_tree : `instance<type<self>>`
            The plugin tree to check.
        
        Returns
        -------
        is_subset : `bool`
        """
        return plugin_tree._plugins.issubset(self._plugins)
    
    
    def has_intersection_with_plugin(self, plugin):
        """
        Returns whether the plugin tree has intersection with the given plugin.
        
        Parameters
        ----------
        plugin : ``Plugin``
            The plugin to check.
        
        Returns
        -------
        has_intersection : `bool`
        """
        return plugin in self._plugins
    
    
    def has_intersection_with_plugins(self, plugins):
        """
        Returns whether the plugin tree has intersection with the given plugins.
        
        Parameters
        ----------
        plugins : `iterable` of ``Plugin``
            The plugins to check.
        
        Returns
        -------
        has_intersection : `bool`
        """
        return (True if self._plugins.intersection(plugins) else False)
    
    
    def get_intersection_with_plugins(self, plugins):
        """
        Returns the plugin tree and given plugins intersection.
        
        Parameters
        ----------
        plugins : `iterable` of ``Plugin``
            The plugins to check.
        
        Returns
        -------
        intersection : `set` of ``Plugin``
        """
        return self._plugins.intersection(plugins)
    
    
    def has_intersection_with_plugin_tree(self, plugin_tree):
        """
        Returns whether the plugin tree has intersection with the given plugin tree.
        
        Parameters
        ----------
        plugin_tree : `instance<type<self>>`
            The plugin tree to check.
        
        Returns
        -------
        has_intersection : `bool`
        """
        return (True if self._plugins.intersection(plugin_tree._plugins) else False)
    
    
    def merge_with_plugin_tree(self, plugin_tree):
        """
        Merges the two plugin trees.
        
        Parameters
        ----------
        plugin_tree : `instance<type<self>>`
            The plugin tree to check.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._plugins = frozenset(chain(self._plugins, plugin_tree._plugins))
        new._plugins_sorted = None
        return new
    
    
    def iter_plugins(self):
        """
        Iterates over the plugins contained by the plugin tree.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin : ``Plugin``
        """
        yield from self._plugins
    
    
    def copy_without(self, plugins):
        """
        Copies the plugin tree without the given plugins.
        
        Parameters
        ----------
        plugins : `set` of ``Plugin``
        
        Returns
        -------
        new : `instance<type<self>>`, `None`
            Returns `None` if the new copy would be empty.
        """
        old_plugins = self._plugins
        new_plugins = old_plugins - plugins
        
        new_length = len(new_plugins)
        if new_length == 0:
            return None
        
        if new_length == len(old_plugins):
            return self
        
        new = object.__new__(type(self))
        new._plugins = new_plugins
        new._plugins_sorted = None
        return new



__all__ = ('CommandBaseCustomId', )

from scarletio import copy_docs

from ...interfaces.command import CommandInterface

from ..command_base import CommandBase


class CommandBaseCustomId(CommandInterface, CommandBase):
    """
    Base type for commands based on `custom_id` matching.
    
    Attributes
    ----------
    _command_function : `async-callable`
        The command's function to call.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _keyword_parameter_converters : `tuple<ParameterConverterBase>`
        Parameter converters for keyword parameters.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        The parent slasher of the component command.
    
    _positional_parameter_converters : `tuple<ParameterConverterBase>`
        Parsers to parse positional command parameters.
    
    _regex_custom_ids : `None | tuple<RegexMatcher>`.
        Regex matchers to match custom-ids.
    
    _string_custom_ids : `None | tuple<str>`
        The custom id-s to wait for.
    
    name : `str`
        The command's name.
        
        Only used for debugging.
    
    response_modifier : `None | ResponseModifier`
        Modifies values returned and yielded to command coroutine processor.
    """
    __slots__ = (
        '_command_function', '_keyword_parameter_converters', '_positional_parameter_converters', '_regex_custom_ids',
        '_string_custom_ids', 'response_modifier',
    )
    
    def __new__(cls, function, name = None, *, custom_id = ..., **keyword_parameters):
        """
        Creates a new `custom_id` based command instance.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        function : `async-callable`
            The function used as the command when using the respective slash command.
        
        name : `None`, `str` = `None`, Optional
            The name of the component command.
        
        custom_id : `str | re.Pattern | (list | set)<str | re.Pattern>`, Optional (Keyword only)
            Custom id to match by the component command.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters
        
        Other Parameters
        ----------------
        allowed_mentions : `None | str, UserBase | Role | AllowedMentionProxy | list<str | UserBase | Role> \
                , Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError
    
    
    @copy_docs(CommandBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts = CommandBase._put_repr_parts_into(self, repr_parts)
            
        string_custom_ids = self._string_custom_ids
        if (string_custom_ids is not None):
            
            repr_parts.append(', string_custom_ids = [')
            index = 0
            limit = len(string_custom_ids)
            
            while True:
                string_custom_id = string_custom_ids[index]
                repr_parts.append(repr(string_custom_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        regex_custom_ids = self._regex_custom_ids
        if (regex_custom_ids is not None):
            
            repr_parts.append(', regex_custom_ids = [')
            index = 0
            limit = len(regex_custom_ids)
            
            while True:
                regex_custom_id = regex_custom_ids[index]
                repr_parts.append(repr(regex_custom_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        response_modifier = self.response_modifier
        if (response_modifier is not None):
            repr_parts.append(', response_modifier = ')
            repr_parts.append(repr(response_modifier))
        
        # _command_function
        repr_parts.append(', command_function = ')
        repr_parts.append(repr(self._command_function))
        
        return repr_parts
    
    
    @copy_docs(CommandBase.__hash__)
    def __hash__(self):
        hash_value = CommandBase.__hash__(self)
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except KeyError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _keyword_parameter_converters
        # Internal field
        
        # _positional_parameter_converters
        # Internal field
        
        # _regex_custom_ids
        regex_custom_ids = self._regex_custom_ids
        if (regex_custom_ids is not None):
            hash_value ^= hash(regex_custom_ids)
        
        # _string_custom_ids
        string_custom_ids = self._string_custom_ids
        if (string_custom_ids is not None):
            hash_value ^= hash(string_custom_ids)
        
        # response_modifier
        response_modifier = self.response_modifier
        if (response_modifier is not None):
            hash_value ^= hash(response_modifier)
        
        return hash_value
    
    
    @copy_docs(CommandBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBase._is_equal_same_type(self, other):
            return False
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _keyword_parameter_converters
        # Internal field
        
        # _positional_parameter_converters
        # Internal field
        
        # _regex_custom_ids
        if self._regex_custom_ids != other._regex_custom_ids:
            return False
        
        # _string_custom_ids
        if self._string_custom_ids != other._string_custom_ids:
            return False
        
        # response_modifier
        if self.response_modifier != other.response_modifier:
            return False
        
        return True
    
    
    async def invoke(self, client, interaction_event, regex_match):
        """
        Calls the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        regex_match : `None | RegexMatch`
            The matched regex if applicable.
        """
        return
    
    
    def copy(self):
        """
        Copies the command.
        
        Returns
        -------
        new : ``CommandBaseCustomId``
        """
        new = CommandBase.copy(self)
        
        # _command_function
        new._command_function = self._command_function
        
        # _keyword_parameter_converters
        new._keyword_parameter_converters = self._keyword_parameter_converters
        
        # _positional_parameter_converters
        new._positional_parameter_converters = self._positional_parameter_converters
        
        # _regex_custom_ids
        new._regex_custom_ids = self._regex_custom_ids
        
        # _string_custom_ids
        new._string_custom_ids = self._string_custom_ids
        
        # response_modifier
        new.response_modifier = self.response_modifier
        
        return new
    
    
    @copy_docs(CommandInterface.get_command_function)
    def get_command_function(self):
        return self._command_function



from .activity import *
from .activity_assets import *
from .activity_field_base import *
from .activity_metadata import *
from .activity_party import *
from .activity_secrets import *
from .activity_timestamps import *


__all__ = (
    *activity.__all__,
    *activity_assets.__all__,
    *activity_field_base.__all__,
    *activity_metadata.__all__,
    *activity_party.__all__,
    *activity_secrets.__all__,
    *activity_timestamps.__all__,
)



__all__ = ('InteractionMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...application_command import ApplicationCommandTargetType
from ...bases import PlaceHolder


class InteractionMetadataBase(RichAttributeErrorBaseType):
    """
    Base type for values assigned to ``InteractionEvent.message`` field.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new interaction metadata from the given parameters.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Construct
        self = object.__new__(cls)
        return self
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates the interaction metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to work with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return cls()
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an new interaction with it's attribute set as it's default values.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)    
    
    
    def copy(self):
        """
        Copies the interaction metadata.
        
        Returns
        -------
        new : `instance<cls>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the interaction metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # Construct
        new = object.__new__(type(self))
        return new
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the interaction metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to work with.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return self.copy_with()
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new interaction from the received data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            The received interaction field data.
        
        guild_id : `int` = `0`, Optional
            The respective guild's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False, guild_id = 0):
        """
        Converts the interaction into a json serailzable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective guild's identifier.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        return {}
    
    
    def __repr__(self):
        """Returns the interaction's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(type(self).__name__)
        
        self._put_attribute_representations_into(repr_parts)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_attribute_representations_into(self, repr_parts):
        """
        Helper function to build representation of the interaction metadata.
        
        Parameters
        ----------
        repr_parts : `list` of `str`
            Integration metadata representation parts.
        
        Returns
        -------
        field_added : `bool`
            Whether any field was added.
        """
        return False
    
    
    def __hash__(self):
        """Returns the interaction's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two interactions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two interactions are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other interaction. Must be the same type as `self`.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    application_command_id = PlaceHolder(
        0,
        """
        The represented application command's identifier number.
        
        Returns
        -------
        application_command_id : `int`
        """,
    )
    
    
    application_command_name = PlaceHolder(
        '',
        """
        The represented application command's name.
        
        Returns
        -------
        application_command_name : `str`
        """,
    )
    
    
    component = PlaceHolder(
        None,
        """
        The interacted component of a message component interaction.
        
        Returns
        -------
        components : ``None | InteractionComponent``
        """
    )
    
    
    components = PlaceHolder(
        None,
        """
        Submitted component values of a form submit interaction.
        
        Returns
        -------
        components : ``None | tuple<InteractionComponent>``
        """
    )
    
    
    custom_id = PlaceHolder(
        None,
        """
        Form interaction's custom identifier.
        
        Returns
        -------
        custom_id : `None | str`
        """,
    )
    
    
    options = PlaceHolder(
        None,
        """
        Application command option representations. Like sub-command or parameter.
            
        Returns
        -------
        options : `None`, `tuple` of ``InteractionOption``
        """,
    )
    
    
    target_id = PlaceHolder(
        0,
        """
        The interaction's target's identifier. Applicable for context commands.
        
        Returns
        -------
        target_id : `int`
        """,
    )
    
    
    target_type = PlaceHolder(
        ApplicationCommandTargetType.none,
        """
        The invoked application command's target type.
        
        Returns
        -------
        target_type : ``ApplicationCommandTargetType``
        """,
    )
    
    # Extra utility
    
    # Application command autocomplete
    
    def iter_options(self):
        """
        Iterates over the options of the interaction application command (autocomplete) interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``InteractionOption``
        """
        options = self.options
        if (options is not None):
            yield from options
    
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction.
        
        Returns
        -------
        option : `None`, ``InteractionOption``
        """
        for option in self.iter_options():
            focused_option = option.focused_option
            if (focused_option is not None):
                return focused_option
    
    
    def get_non_focused_values(self):
        """
        Gets the non focused values of the interaction.
        
        Returns
        -------
        non_focused_options : `dict<str, None | str>`
        """
        return dict(self._iter_non_focused_values())
    
    
    def _iter_non_focused_values(self):
        """
        Iterates over the non focused values of the interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            The option's name.
        
        value : `None | str`
            The option's value.
        """
        for option in self.iter_options():
            yield from option._iter_non_focused_values()
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None | str`
            The value, the user has been typed.
        """
        if not option_names:
            return
            
        option_name, *option_names = option_names
        
        for option in self.iter_options():
            if option.name == option_name:
                return option.get_value_of(*option_names)
    
    # Form submit
    
    def iter_components(self):
        """
        Iterates over the sub-components of a form-submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``InteractionComponent``
        """
        return
        yield
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id_is_multi_value_values : `(str, ComponentType, None | str | tuple<str>)`
        """
        return
        yield



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory,
    force_string_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, default_entity_putter_factory, entity_id_putter_factory, force_string_putter_factory,
    optional_entity_id_array_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator_factory, entity_id_array_validator_factory,
    entity_id_validator_factory
)
from ...role import Role
from ...user import ClientUserBase, User, ZEROUSER

from .constants import NAME_ALLOWED_CHARACTERS, NAME_LENGTH_MAX, NAME_LENGTH_MIN

# animated

parse_animated = bool_parser_factory('animated', False)
put_animated = bool_optional_putter_factory('animated', False)
validate_animated = bool_validator_factory('animated', False)

# available

parse_available = bool_parser_factory('available', True)
put_available = bool_optional_putter_factory('available', True)
validate_available = bool_validator_factory('available', True)

# guild_id

validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('emoji_id')

# managed

parse_managed = bool_parser_factory('managed', False)
put_managed = bool_optional_putter_factory('managed', False)
validate_managed = bool_validator_factory('managed', False)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')


def validate_name(name):
    """
    Validates the given emoji name.
    
    Parameters
    ----------
    name : `None`, `str`
        Emoji name to validate.
    
    Returns
    -------
    name : `str`
        The validated name.
    
    Raises
    ------
    TypeError
        - If `name`'s type is incorrect.
    ValueError
        - if `name`'s value is incorrect.
    """
    if name is None:
        return ''
    
    if not isinstance(name, str):
        raise TypeError(
            f'`name` can be `None`, `str`, got {name.__class__.__name__} ;{name!r}.'
        )
    
    if not name:
        return ''
    
    validated_name = ''.join(NAME_ALLOWED_CHARACTERS.findall(name))
    
    name_length = len(validated_name)
    if (name_length < NAME_LENGTH_MIN) or (name_length > NAME_LENGTH_MAX):
        raise ValueError(
            f'`name` length can be >= {NAME_LENGTH_MIN} and <= {NAME_LENGTH_MAX}, got {name_length}; '
            f'validated_name = {validated_name}; input_name = {name!r}'
        )
    
    return validated_name


# require_colons

parse_require_colons = bool_parser_factory('require_colons', True)
put_require_colons = bool_optional_putter_factory('require_colons', True)
validate_require_colons = bool_validator_factory('require_colons', True)

# role_ids

parse_role_ids = entity_id_array_parser_factory('roles')
put_role_ids = optional_entity_id_array_optional_putter_factory('roles')
validate_role_ids = entity_id_array_validator_factory('role_ids', Role)

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user = default_entity_putter_factory('user', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)



__all__ = ()

from math import ceil

from ...channel import Channel
from ...field_parsers import entity_id_parser_factory, int_parser_factory, nullable_string_parser_factory
from ...field_putters import entity_id_optional_putter_factory, int_putter_factory, nullable_string_putter_factory
from ...field_validators import entity_id_validator_factory, nullable_string_validator_factory

from .constants import AUTO_MODERATION_ACTION_TIMEOUT_MAX, CUSTOM_MESSAGE_LENGTH_MAX

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_optional_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# custom_message

parse_custom_message = nullable_string_parser_factory('custom_message')
put_custom_message = nullable_string_putter_factory('custom_message')
validate_custom_message = nullable_string_validator_factory('custom_message', 0, CUSTOM_MESSAGE_LENGTH_MAX)

# duration

parse_duration = int_parser_factory('duration_seconds', 0)
put_duration = int_putter_factory('duration_seconds')

def validate_duration(duration):
    """
    Validates the given duration.
    
    Parameters
    ----------
    duration : `None | int | float`
        The timeout's duration applied on trigger.
    
    Returns
    -------
    duration : `int`
    
    Raises
    ------
    TypeError
        - If `duration` type is incorrect.
    """
    if duration is None:
        duration = 0
    
    elif isinstance(duration, int):
        pass
    
    elif isinstance(duration, float):
        duration = ceil(duration)
    
    else:
        raise TypeError(
            f'`duration` can be `None | int | float`, got {type(duration).__name__}; {duration!r}.'
        )
    
    if duration < 0:
        duration = 0
    
    elif duration > AUTO_MODERATION_ACTION_TIMEOUT_MAX:
        duration = AUTO_MODERATION_ACTION_TIMEOUT_MAX
    
    return duration



__all__ = ('SlashCommand',)

from scarletio import copy_docs, export

from .....discord.application import ApplicationIntegrationType
from .....discord.application_command import ApplicationCommandTargetType, INTEGRATION_CONTEXT_TYPES_ALL
from .....discord.application_command.application_command.constants import (
    OPTIONS_MAX as APPLICATION_COMMAND_OPTIONS_MAX
)
from .....discord.events.handling_helpers import check_name
from .....discord.permission import Permission

from ...constants import (
    APPLICATION_COMMAND_CATEGORY_DEEPNESS_START, APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND,
    APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY
)
from ...converters import get_slash_command_parameter_converters
from ...exceptions import SlashCommandParameterConversionError, handle_command_exception
from ...interfaces.autocomplete import AutocompleteInterface
from ...interfaces.command import CommandInterface
from ...interfaces.nestable import NestableInterface
from ...interfaces.self_reference import SelfReferenceInterface
from ...response_modifier import ResponseModifier
from ...utils import (
    UNLOADING_BEHAVIOUR_DELETE, UNLOADING_BEHAVIOUR_INHERIT, UNLOADING_BEHAVIOUR_KEEP, raw_name_to_display
)
from ...wrappers import CommandWrapper, get_parameter_configurers

from ..command_base_application_command import CommandBaseApplicationCommand
from ..command_base_application_command.helpers import (
    _reset_application_command_schema, _validate_delete_on_unload, _validate_guild,
    _validate_integration_context_types, _validate_integration_types, _validate_is_global, _validate_name,
    _validate_nsfw, _validate_required_permissions
)

from .helpers import _generate_description_from, _validate_default
from .slash_command_category import SlashCommandCategory
from .slash_command_function import SlashCommandFunction
from .slash_command_parameter_auto_completer import SlashCommandParameterAutoCompleter


@export
class SlashCommand(
    AutocompleteInterface, CommandInterface, NestableInterface, SelfReferenceInterface, CommandBaseApplicationCommand
):
    """
    Class to wrap an application command providing interface for ``Slasher``.
    
    Attributes
    ----------
    _auto_completers : `None | list<SlashCommandParameterAutoCompleter>`
        Auto completer functions.
    
    _command : `None | SlashCommandFunction`
        The command of the slash command.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        Reference to the slasher application command's parent.
    
    _permission_overwrites : `None | dict<int, list<ApplicationCommandPermissionOverwrite>>
        Permission overwrites applied to the slash command.
    
    _registered_application_command_ids : `None | dict<int, int>`
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None | ApplicationCommand`
        Internal slot used by the ``.get_schema`` method.
    
    _self_reference : `None | WeakReferer<instance>`
        Back reference to the slasher application command.
        
        Used by sub commands to access the parent entity.
    
    _sub_commands : `None | dict<str, SlashCommandFunction | SlashCommandCategory>`
        Sub-commands of the slash command.
        
        Mutually exclusive with the ``._command`` parameter.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    default : `bool`
        Whether the command is the default command in it's category.
    
    description : `str`
        Application command description. It's length can be in range [2:100].
    
    global_ : `bool`
        Whether the command is a global command.
        
        Global commands have their ``.guild_ids`` set as `None`.
    
    guild_ids : `None | set<int>`
        The ``Guild``'s id to which the command is bound to.
    
    integration_context_types : `None | tuple<ApplicationCommandIntegrationContextType>`
        The places where the application command shows up.
    
    integration_types : `None | tuple<ApplicationIntegrationType>`
        The options where the application command can be integrated to.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    nsfw : `None | bool`
        Whether the application command is only allowed in nsfw channels.
    
    required_permissions : ``Permission``
        The required permissions to use the application command inside of a guild.
    
    Notes
    -----
    Slash commands are weakreferable.
    """
    __slots__ = (
        '__weakref__', '_auto_completers', '_command', '_self_reference', '_sub_commands', 'default', 'description'
    )
    
    
    def __new__(
        cls,
        function,
        name = None,
        *,
        default = ...,
        delete_on_unload = ...,
        description = ...,
        guild = ...,
        integration_context_types = ...,
        integration_types = ...,
        is_global = ...,
        nsfw = ...,
        required_permissions = ...,
        **keyword_parameters,
    ):
        """
        Creates a new ``SlashCommand`` with the given parameters.
        
        Parameters
        ----------
        function : `None | async-callable`
            The function used as the command when using the respective slash command.
        
        name : `None | str` = `None`, Optional
            The command's name if applicable. If not given or if given as `None`, the `func`'s name will be use
            instead.
        
        default : `bool`, Optional (Keyword only)
            Whether the context command is the default command in it's category.
        
        delete_on_unload : `bool`, Optional (Keyword only)
            Whether the command should be deleted from Discord when removed.
        
        description : `str | object`, Optional (Keyword only)
            Description to use instead of the function's docstring.
        
        guild : ``int | Guild | (list | set)<int | Guild>``, Optional (Keyword only)
            To which guild(s) the command is bound to.
        
        integration_context_types : `None | ApplicationCommandIntegrationContextType | int | str | \
                iterable<ApplicationCommandIntegrationContextType | int | str>`, Optional (Keyword only)
            The places where the application command shows up.
        
        integration_types : `None | ApplicationIntegrationType | int | str | \
                iterable<ApplicationIntegrationType | int | str>`, Optional
            The options where the application command can be integrated to.
        
        is_global : `None | bool` = `None`, Optional
            Whether the slash command is the default command in it's category.
        
        nsfw : `None`, Optional (Keyword only)
            Whether the application command is only allowed in nsfw channels.
        
        required_permissions : `int | Permission`, Optional (Keyword only)
            The required permissions to use the application command inside of a guild.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        allowed_mentions : `None | str, UserBase | Role | AllowedMentionProxy | list<str | UserBase | Role> \
                , Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if (function is not None) and isinstance(function, CommandWrapper):
            command_function, wrappers = function.fetch_function_and_wrappers_back()
        else:
            command_function = function
            wrappers = None
        
        # Pre validate
        name = _validate_name(name)
        
        # delete_on_unload
        if delete_on_unload is ...:
            unloading_behaviour = UNLOADING_BEHAVIOUR_INHERIT
        elif _validate_delete_on_unload(delete_on_unload):
            unloading_behaviour = UNLOADING_BEHAVIOUR_DELETE
        else:
            unloading_behaviour = UNLOADING_BEHAVIOUR_KEEP
        
        # description (no pre-validation)
        if description is ...:
            description = None
        
        # guild
        if guild is ...:
            guild_ids = None
        else:
            guild_ids = _validate_guild(guild)
        
        # integration_context_types
        if integration_context_types is ...:
            integration_context_types = INTEGRATION_CONTEXT_TYPES_ALL
        else:
            integration_context_types = _validate_integration_context_types(integration_context_types)
        
        # integration_types
        if integration_types is ...:
            integration_types = (ApplicationIntegrationType.guild_install,)
        else:
            integration_types = _validate_integration_types(integration_types)
        
        # default
        if default is ...:
            default = False
        else:
            default = _validate_default(default)
        
        # is_global
        if is_global is ...:
            is_global = False
        else:
            is_global = _validate_is_global(is_global)
        
        # nsfw
        if nsfw is ...:
            nsfw = False
        else:
            nsfw = _validate_nsfw(nsfw)
        
        # required_permissions
        if required_permissions is ...:
            required_permissions = Permission()
        else:
            required_permissions = _validate_required_permissions(required_permissions)
        
        # Check extra parameters
        response_modifier = ResponseModifier(keyword_parameters)
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        # Post validate
        name = check_name(command_function, name)
        name = raw_name_to_display(name)
        
        description = _generate_description_from(command_function, name, description)
        
        if command_function is None:
            parameter_converters = None
        else:
            parameter_configurers = get_parameter_configurers(wrappers)
            command_function, parameter_converters = get_slash_command_parameter_converters(command_function, parameter_configurers)
        
        if is_global and (guild_ids is not None):
            raise TypeError(
                f'`is_global` and `guild` contradict each other, got is_global = {is_global!r}, '
                f'guild = {guild!r}.'
            )
        
        if not is_global:
            integration_types = None
            integration_context_types = None
        
        if (command_function is None):
            slash_command_function = None
        else:
            slash_command_function = SlashCommandFunction(
                command_function, parameter_converters, name, description, response_modifier, default
            )
        
        # Construct
        self = object.__new__(cls)
        self._command = slash_command_function
        self._sub_commands = None
        self.description = description
        self.guild_ids = guild_ids
        self.global_ = is_global
        self.name = name
        self._schema = None
        self._registered_application_command_ids = None
        self.default = default
        self._unloading_behaviour = unloading_behaviour
        self.nsfw = nsfw
        self.required_permissions = required_permissions
        self._permission_overwrites = None
        self._auto_completers = None
        self._exception_handlers = None
        self._parent_reference = None
        self._self_reference = None
        self.integration_context_types = integration_context_types
        self.integration_types = integration_types
        
        if (slash_command_function is not None):
            slash_command_function._parent_reference = self.get_self_reference()
        
        if (wrappers is not None):
            for wrapper in wrappers:
                wrapper.apply(self)
        
        return self
    
    
    @copy_docs(CommandBaseApplicationCommand._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts = CommandBaseApplicationCommand._put_repr_parts_into(self, repr_parts)
        
        # default
        default = self.default
        if default:
            repr_parts.append(', default = ')
            repr_parts.append(repr(default))
        
        # description
        description = self.description
        if self.name != description:
             repr_parts.append(', description = ')
             repr_parts.append(repr(description))
        
        # _command
        command = self._command
        if (command is not None):
            repr_parts.append(', command = ')
            repr_parts.append(repr(command))
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            repr_parts.append(', sub_commands = ')
            repr_parts.append(repr(sub_commands))
        
        return repr_parts
    
    
    @copy_docs(CommandBaseApplicationCommand.__hash__)
    def __hash__(self):
        hash_value = CommandBaseApplicationCommand.__hash__(self)
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            hash_value ^= len(auto_completers) << 22
            
            for auto_completer in auto_completers:
                hash_value ^= hash(auto_completer)
        
        # _command
        command = self._command
        if (command is not None):
            hash_value ^= hash(command)
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            hash_value ^= len(sub_commands) << 26
            
            for sub_command in sub_commands:
                hash_value ^= hash(sub_command)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        
        return hash_value
    
    
    @copy_docs(CommandBaseApplicationCommand._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBaseApplicationCommand._is_equal_same_type(self, other):
            return False
        
        # _auto_completers
        if self._auto_completers != other._auto_completers:
            return False
        
        # _command
        if self._command != other._command:
            return False
        
        # _self_reference
        # Internal field
        
        # _sub_commands
        if self._sub_commands != other._sub_commands:
            return False
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        return True
    
    
    @copy_docs(CommandBaseApplicationCommand.invoke)
    async def invoke(self, client, interaction_event):
        options = interaction_event.options
        
        command = self._command
        if (command is not None):
            await command.invoke(client, interaction_event, options)
            return
        
        sub_commands = self._sub_commands
        if (sub_commands is None):
            return
        
        if (options is None) or (len(options) != 1):
            return
        
        option = options[0]
        
        try:
            sub_command = sub_commands[option.name]
        except KeyError:
            pass
        else:
            await sub_command.invoke(client, interaction_event, option.options)
            return
        
        # Do not put this into the `except` branch.
        await handle_command_exception(
            self,
            client,
            interaction_event,
            SlashCommandParameterConversionError(
                None,
                option.name,
                'sub-command',
                [*sub_commands.keys()],
            )
        )
        return
    
    
    @copy_docs(CommandBaseApplicationCommand.copy)
    def copy(self):
        new = CommandBaseApplicationCommand.copy(self)
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            auto_completers = auto_completers.copy()
        new._auto_completers = auto_completers
        
        # _command
        command = self._command
        if (command is not None):
            command = command.copy()
        new._command = command
        
        # _self_reference
        new._self_reference = None
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            sub_commands = {name: category.copy() for name, category in sub_commands.items()}
        new._sub_commands = sub_commands
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # ---- POST LINKING ----
        
        if (sub_commands is not None):
            for sub_command in sub_commands.values():
                sub_command._parent_reference = new.get_self_reference()
        
        return new
    
    
    @CommandBaseApplicationCommand.target_type.getter
    def target_type(self):
        return ApplicationCommandTargetType.chat

    
    @copy_docs(CommandInterface.get_command_function)
    def get_command_function(self):
        slash_command_function = self._command
        if (slash_command_function is not None):
            return slash_command_function.get_command_function()
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the respective auto completion function of the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        
        auto_complete_option : `InteractionMetadataApplicationCommandAutocomplete | InteractionOption`
            The option to autocomplete.
        """
        command_function = self._command
        if (command_function is not None):
            await command_function.invoke_auto_completion(client, interaction_event, auto_complete_option)
            return
        
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            auto_complete_option = auto_complete_option.options[0]
            
            auto_complete_option_type = auto_complete_option.type
            if (
                (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND) or
                (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY)
            ):
                options = auto_complete_option.options
                if (options is not None):
                    try:
                        sub_command = sub_commands[auto_complete_option.name]
                    except KeyError:
                        pass
                    else:
                        await sub_command.invoke_auto_completion(client, interaction_event, auto_complete_option)
            
            return
        
        # no more cases
    
    @copy_docs(CommandBaseApplicationCommand._get_schema_options)
    def _get_schema_options(self):
        while True:
            command = self._command
            if (command is not None):
                parameter_converters = command._parameter_converters
                
                options = None
                for parameter_converter in parameter_converters:
                    option = parameter_converter.as_option()
                    if (option is not None):
                        if (options is None):
                            options = []
                        
                        options.append(option)
                break
            
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                options = [sub_command.as_option() for name, sub_command in sorted(sub_commands.items())]
                break
            
            options = None
            break
        
        return options
    
    
    def as_sub_command(self, deepness):
        """
        Returns the slash command as a sub-command or sub-category.
        
        Parameters
        ----------
        deepness : `int`
            How nested the category or function will be.
        
        Returns
        -------
        new : `SlashCommandFunction | SlashCommandCategory`
        """
        command = self._command
        if (command is not None):
            return command
        
        return SlashCommandCategory(self.name, self.description, self.default, deepness)
    
    
    @copy_docs(NestableInterface._is_nestable)
    def _is_nestable(self):
        return True if self.get_command_function() is None else False
    
    
    @copy_docs(NestableInterface._make_command_instance_from_parameters)
    def _make_command_instance_from_parameters(self, function, positional_parameters, keyword_parameters):
        return type(self)(function, *positional_parameters, **keyword_parameters)
    
    
    @copy_docs(NestableInterface._store_command_instance)
    def _store_command_instance(self, command):
        if isinstance(command, type(self)):
            instance = self._add_application_command(command)
            return True, instance
        
        return False, None
        
    
    def _add_application_command(self, command):
        """
        Adds a sub-command or sub-category to the slash command.
        
        Parameters
        ----------
        command : ``SlashCommand``
            The slash command to add.
        
        Raises
        ------
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        sub_commands = self._sub_commands
        if (
            (sub_commands is not None) and
            (len(sub_commands) == APPLICATION_COMMAND_OPTIONS_MAX) and
            (command.name not in sub_commands)
        ):
            raise RuntimeError(
                f'The {self!r} reached the maximal amount of children '
                f'({APPLICATION_COMMAND_OPTIONS_MAX}).'
            )
        
        if (sub_commands is not None) and command.default:
            for sub_command in sub_commands.values():
                if sub_command.default and (sub_command.name != command.name):
                    raise RuntimeError(
                        f'{self!r} already has a default command.'
                    )
        
        as_sub_command = command.as_sub_command(APPLICATION_COMMAND_CATEGORY_DEEPNESS_START)
        as_sub_command._parent_reference = self.get_self_reference()
        
        if sub_commands is None:
            sub_commands = {}
            self._sub_commands = sub_commands
        
        sub_commands[as_sub_command.name] = as_sub_command
        _reset_application_command_schema(self)
        
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            for auto_completer in auto_completers:
                as_sub_command._try_resolve_auto_completer(auto_completer)
        
        return as_sub_command
    
    
    @copy_docs(CommandBaseApplicationCommand.get_real_command_count)
    def get_real_command_count(self):
        while True:
            if (self._command is not None):
                real_command_count = 1
                break
            
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                real_command_count = 0
                
                for sub_command_or_category in sub_commands.values():
                    if isinstance(sub_command_or_category, SlashCommandFunction):
                        real_command_count += 1
                        continue
                    
                    # Nesting more is not allowed by Discord.
                    sub_commands = sub_command_or_category._sub_commands
                    if (sub_commands is not None):
                        real_command_count += len(sub_commands)
                break
            
            real_command_count = 0
            break
        
        return real_command_count
    
    
    @copy_docs(AutocompleteInterface._register_auto_completer)
    def _register_auto_completer(self, function, parameter_names):
        while True:
            slash_command_function = self._command
            if (slash_command_function is not None):
                auto_completer = slash_command_function._register_auto_completer(function, parameter_names)
                break
                
            auto_completer = self._make_auto_completer(function, parameter_names)
            self._store_auto_completer(auto_completer)
            
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                for sub_command in sub_commands.values():
                    sub_command._try_resolve_auto_completer(auto_completer)
                break
            
            break
        
        _reset_application_command_schema(self)
        return auto_completer
    
    
    def _try_resolve_auto_completer(self, auto_completer):
        """
        Tries to register auto completer to the slasher application command.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer.
        
        Returns
        -------
        resolved : `int`
            The amount of parameters resolved.
        """
        resolved = 0
        
        command = self._command
        if command is None:
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                for sub_command in sub_commands.values():
                    resolved += sub_command._try_resolve_auto_completer(auto_completer)
        else:
            resolved += command._try_resolve_auto_completer(auto_completer)
        
        if resolved:
            _reset_application_command_schema(self)
        
        return resolved
